from ._functions import (
    list_ml_models,
    create_ml_model,
    delete_ml_model,
    activate_ml_model_endpoint_version,
    deactivate_all_ml_model_endpoint_versions,
    deactivate_ml_model_endpoint_version,
    list_ml_model_endpoint_versions,
    score_ml_model_endpoint,
    score_ml_model_endpoint_version,
)

__all__ = [
    "list_ml_models",
    "create_ml_model",
    "delete_ml_model",
    "activate_ml_model_endpoint_version",
    "deactivate_all_ml_model_endpoint_versions",
    "deactivate_ml_model_endpoint_version",
    "list_ml_model_endpoint_versions",
    "score_ml_model_endpoint",
    "score_ml_model_endpoint_version",
]



from typing import Optional
import pandas as pd
import datetime
from sempy._utils._log import log
from ._reportwrapper import connect_report
from ._report_bpa_rules import report_bpa_rules
from .._helper_functions import (
    format_dax_object_name,
    save_as_delta_table,
    resolve_item_name_and_id,
    resolve_workspace_capacity,
    _get_column_aggregate,
    resolve_workspace_name_and_id,
)
from ..lakehouse import get_lakehouse_tables, lakehouse_attached
import sempy_labs._icons as icons
from IPython.display import display, HTML
from uuid import UUID


@log
def run_report_bpa(
    report: str,
    rules: Optional[pd.DataFrame] = None,
    workspace: Optional[str | UUID] = None,
    # language: Optional[str] = None,
    export: bool = False,
    return_dataframe: bool = False,
):
    """
    Displays an HTML visualization of the results of the Best Practice Analyzer scan for a report.
    Note: As with all functions which rely on the ReportWrapper, this function requires the report to be in the 'PBIR' format.

    Parameters
    ----------
    report : str
        Name of the report.
    rules : pandas.DataFrame, default=None
        A pandas dataframe containing rules to be evaluated.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    export : bool, default=False
        If True, exports the resulting dataframe to a delta table in the lakehouse attached to the notebook.
    return_dataframe : bool, default=False
        If True, returns a pandas dataframe instead of the visualization.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe in HTML format showing report objects which violated the best practice analyzer rules.
    """

    with connect_report(
        report=report, workspace=workspace, readonly=True, show_diffs=False
    ) as rpt:

        dfCV = rpt.list_custom_visuals()
        dfP = rpt.list_pages()
        dfRF = rpt.list_report_filters()
        dfRF["Filter Object"] = format_dax_object_name(
            dfRF["Table Name"], dfRF["Object Name"]
        )
        dfPF = rpt.list_page_filters()

        dfPF["Filter Object"] = (
            dfPF["Page Display Name"]
            + " : "
            + format_dax_object_name(dfPF["Table Name"], dfPF["Object Name"])
        )
        dfVF = rpt.list_visual_filters()
        dfVF["Filter Object"] = (
            format_dax_object_name(dfVF["Page Display Name"], dfVF["Visual Name"])
            + " : "
            + format_dax_object_name(dfVF["Table Name"], dfVF["Object Name"])
        )
        dfRLM = rpt.list_report_level_measures()
        dfV = rpt.list_visuals()
        dfV["Visual Full Name"] = format_dax_object_name(
            dfV["Page Display Name"], dfV["Visual Name"]
        )
        dfSMO = rpt.list_semantic_model_objects(extended=True)

        # Translations
        if rules is None:
            rules = report_bpa_rules()

        scope_to_dataframe = {
            "Custom Visual": (dfCV, ["Custom Visual Display Name"]),
            "Page": (dfP, ["Page Display Name"]),
            "Visual": (dfV, ["Visual Full Name"]),
            "Report Filter": (dfRF, ["Filter Object"]),
            "Page Filter": (dfPF, ["Filter Object"]),
            "Visual Filter": (dfVF, ["Filter Object"]),
            "Report Level Measure": (dfRLM, ["Measure Name"]),
            "Semantic Model": (dfSMO, ["Report Source Object"]),
        }

        def execute_rule(row):
            scopes = row["Scope"]

            # support both str and list as scope type
            if isinstance(scopes, str):
                scopes = [scopes]

            # collect output dataframes
            df_outputs = []

            for scope in scopes:
                # common fields for each scope
                (df, violation_cols_or_func) = scope_to_dataframe[scope]

                # execute rule and subset df
                df_violations = df[row["Expression"](df)]

                # subset the right output columns (e.g. Table Name & Column Name)
                if isinstance(violation_cols_or_func, list):
                    violation_func = lambda violations: violations[
                        violation_cols_or_func
                    ]
                else:
                    violation_func = violation_cols_or_func

                # build output data frame
                df_output = violation_func(df_violations).copy()

                df_output.columns = ["Object Name"]
                df_output["Rule Name"] = row["Rule Name"]
                df_output["Category"] = row["Category"]

                # Handle 'Object Type' based on scope
                if scope == "Semantic Model":
                    # Ensure 'Report Source Object' is unique in dfSMO
                    dfSMO_unique = dfSMO.drop_duplicates(subset="Report Source Object")
                    # Map 'Object Name' to the 'Report Source' column in dfSMO
                    df_output["Object Type"] = df_output["Object Name"].map(
                        dfSMO_unique.set_index("Report Source Object")["Report Source"]
                    )
                else:
                    df_output["Object Type"] = scope

                df_output["Severity"] = row["Severity"]
                df_output["Description"] = row["Description"]
                if scope in ["Page", "Page Filter"]:
                    if not df_output.empty:
                        for idx, r in df_output.iterrows():
                            df_output.loc[idx, "URL"] = rpt._get_url(
                                page_name=r["Object Name"]
                            )
                elif scope in ["Visual", "Visual Filter"]:
                    if not df_output.empty:
                        for idx, r in df_output.iterrows():
                            df_output.loc[idx, "URL"] = rpt._get_url(
                                page_name=r["Object Name"].split("[")[0][1:-1],
                                visual_name=r["Object Name"].split("[")[1].rstrip("]"),
                            )
                else:
                    df_output["URL"] = rpt._get_url()
                df_outputs.append(df_output)

            return df_outputs

    # flatten list of lists
    flatten_dfs = [
        df for dfs in rules.apply(execute_rule, axis=1).tolist() for df in dfs
    ]

    finalDF = pd.concat(flatten_dfs, ignore_index=True)
    finalDF = finalDF[
        [
            "Category",
            "Rule Name",
            "Object Type",
            "Object Name",
            "Severity",
            "Description",
            "URL",
            # "Report URL",
        ]
    ]

    if return_dataframe:
        return finalDF

    if export:
        if not lakehouse_attached():
            raise ValueError(
                f"{icons.red_dot} In order to export the BPA results, a lakehouse must be attached to the notebook."
            )

        now = datetime.datetime.now()
        delta_table_name = "reportbparesults"
        lakeT = get_lakehouse_tables()
        lakeT_filt = lakeT[lakeT["Table Name"] == delta_table_name]

        if len(lakeT_filt) == 0:
            runId = 1
        else:
            max_run_id = _get_column_aggregate(table_name=delta_table_name)
            runId = max_run_id + 1

        (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
        (report_name, report_id) = resolve_item_name_and_id(
            item=report, type="Report", workspace=workspace_id
        )

        export_df = finalDF.copy()
        capacity_id, capacity_name = resolve_workspace_capacity(workspace=workspace_id)
        export_df["Capacity Name"] = capacity_name
        export_df["Capacity Id"] = capacity_id
        export_df["Workspace Name"] = workspace_name
        export_df["Workspace Id"] = workspace_id
        export_df["Report Name"] = report_name
        export_df["Report Id"] = report_id
        export_df["RunId"] = runId
        export_df["Timestamp"] = now
        export_df["RunId"] = export_df["RunId"].astype(int)

        export_df = export_df[
            [
                "Capacity Name",
                "Capacity Id",
                "Workspace Name",
                "Workspace Id",
                "Report Name",
                "Report Id",
                "Category",
                "Rule Name",
                "Object Type",
                "Object Name",
                "Severity",
                "Description",
                "URL",
            ]
        ]

        save_as_delta_table(
            dataframe=export_df,
            delta_table_name=delta_table_name,
            write_mode="append",
            merge_schema=True,
        )

        return

    finalDF.replace(
        {"Warning": icons.warning, "Error": icons.error, "Info": icons.info},
        inplace=True,
    )

    pd.set_option("display.max_colwidth", 100)

    finalDF = (
        finalDF[
            [
                "Category",
                "Rule Name",
                "Object Type",
                "Object Name",
                "Severity",
                "Description",
                "URL",
                # "Report URL",
            ]
        ]
        .sort_values(["Category", "Rule Name", "Object Type", "Object Name"])
        .set_index(["Category", "Rule Name"])
    )

    bpa2 = finalDF.reset_index()
    bpa_dict = {
        cat: bpa2[bpa2["Category"] == cat].drop("Category", axis=1)
        for cat in bpa2["Category"].drop_duplicates().values
    }

    styles = """
    <style>
        .tab { overflow: hidden; border: 1px solid #ccc; background-color: #f1f1f1; }
        .tab button { background-color: inherit; float: left; border: none; outline: none; cursor: pointer; padding: 14px 16px; transition: 0.3s; }
        .tab button:hover { background-color: #ddd; }
        .tab button.active { background-color: #ccc; }
        .tabcontent { display: none; padding: 6px 12px; border: 1px solid #ccc; border-top: none; }
        .tabcontent.active { display: block; }
        .tooltip { position: relative; display: inline-block; }
        .tooltip .tooltiptext { visibility: hidden; width: 300px; background-color: #555; color: #fff; text-align: center; border-radius: 6px; padding: 5px; position: absolute; z-index: 1; bottom: 125%; left: 50%; margin-left: -110px; opacity: 0; transition: opacity 0.3s; }
        .tooltip:hover .tooltiptext { visibility: visible; opacity: 1; }
    </style>
    """

    # JavaScript for tab functionality
    script = """
    <script>
    function openTab(evt, tabName) {
        var i, tabcontent, tablinks;
        tabcontent = document.getElementsByClassName("tabcontent");
        for (i = 0; i < tabcontent.length; i++) {
            tabcontent[i].style.display = "none";
        }
        tablinks = document.getElementsByClassName("tablinks");
        for (i = 0; i < tablinks.length; i++) {
            tablinks[i].className = tablinks[i].className.replace(" active", "");
        }
        document.getElementById(tabName).style.display = "block";
        evt.currentTarget.className += " active";
    }
    </script>
    """

    # HTML for tabs
    tab_html = '<div class="tab">'
    content_html = ""
    for i, (title, df) in enumerate(bpa_dict.items()):
        if df.shape[0] == 0:
            continue

        tab_id = f"tab{i}"
        active_class = ""
        if i == 0:
            active_class = "active"

        summary = " + ".join(
            [f"{idx} ({v})" for idx, v in df["Severity"].value_counts().items()]
        )
        tab_html += f'<button class="tablinks {active_class}" onclick="openTab(event, \'{tab_id}\')"><b>{title}</b><br/>{summary}</button>'
        content_html += f'<div id="{tab_id}" class="tabcontent {active_class}">'

        # Adding tooltip for Rule Name using Description column
        content_html += '<table border="1">'
        content_html += "<tr><th>Rule Name</th><th>Object Type</th><th>Object Name</th><th>Severity</th></tr>"
        for _, row in df.iterrows():
            content_html += "<tr>"
            if pd.notnull(row["URL"]):
                content_html += f'<td class="tooltip" onmouseover="adjustTooltipPosition(event)"><a href="{row["URL"]}">{row["Rule Name"]}</a><span class="tooltiptext">{row["Description"]}</span></td>'
            elif pd.notnull(row["Description"]):
                content_html += f'<td class="tooltip" onmouseover="adjustTooltipPosition(event)">{row["Rule Name"]}<span class="tooltiptext">{row["Description"]}</span></td>'
            else:
                content_html += f'<td>{row["Rule Name"]}</td>'
            content_html += f'<td>{row["Object Type"]}</td>'
            if pd.notnull(row["URL"]):
                content_html += (
                    f'<td><a href="{row["URL"]}">{row["Object Name"]}</a></td>'
                )
            else:
                content_html += f'<td>{row["Object Name"]}</td>'
            content_html += f'<td style="text-align: center;">{row["Severity"]}</td>'
            content_html += "</tr>"
        content_html += "</table>"

        content_html += "</div>"
    tab_html += "</div>"

    # Display the tabs, tab contents, and run the script
    return display(HTML(styles + tab_html + content_html + script))



import pandas as pd
import re
from .._helper_functions import (
    format_dax_object_name,
    resolve_workspace_name_and_id,
    resolve_lakehouse_name_and_id,
    _create_dataframe,
    _get_delta_table,
    _pure_python_notebook,
)
from typing import Optional
from sempy._utils._log import log
from uuid import UUID
import sempy_labs._icons as icons


@log
def get_lakehouse_columns(
    lakehouse: Optional[str | UUID] = None, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Shows the tables and columns of a lakehouse and their respective properties. This function can be executed in either a PySpark or pure Python notebook. Note that data types may show differently when using PySpark vs pure Python.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    lakehouse : str | uuid.UUID, default=None
        The Fabric lakehouse name or ID.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    lakehouse_workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID used by the lakehouse.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        Shows the tables/columns within a lakehouse and their properties.
    """
    from ._get_lakehouse_tables import get_lakehouse_tables

    columns = {
        "Workspace Name": "string",
        "Lakehouse Name": "string",
        "Table Name": "string",
        "Column Name": "string",
        "Full Column Name": "string",
        "Data Type": "string",
    }
    df = _create_dataframe(columns=columns)

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (lakehouse_name, lakehouse_id) = resolve_lakehouse_name_and_id(
        lakehouse=lakehouse, workspace=workspace_id
    )

    tables = get_lakehouse_tables(
        lakehouse=lakehouse_id, workspace=workspace_id, extended=False, count_rows=False
    )
    tables_filt = tables[tables["Format"] == "delta"]

    def add_column_metadata(table_name, col_name, data_type):
        new_rows.append(
            {
                "Workspace Name": workspace_name,
                "Lakehouse Name": lakehouse_name,
                "Table Name": table_name,
                "Column Name": col_name,
                "Full Column Name": format_dax_object_name(table_name, col_name),
                "Data Type": data_type,
            }
        )

    new_rows = []

    for _, r in tables_filt.iterrows():
        table_name = r["Table Name"]
        path = r["Location"]

        if _pure_python_notebook():
            from deltalake import DeltaTable

            table_schema = DeltaTable(path).schema()

            for field in table_schema.fields:
                col_name = field.name
                match = re.search(r'"(.*?)"', str(field.type))
                if not match:
                    raise ValueError(
                        f"{icons.red_dot} Could not find data type for column {col_name}."
                    )
                data_type = match.group(1)
                add_column_metadata(table_name, col_name, data_type)
        else:
            delta_table = _get_delta_table(path=path)
            table_df = delta_table.toDF()

            for col_name, data_type in table_df.dtypes:
                add_column_metadata(table_name, col_name, data_type)

    return pd.concat([df, pd.DataFrame(new_rows)], ignore_index=True)



from typing import Optional, List
from uuid import UUID
from sempy.fabric.exceptions import FabricHTTPException
import time
import sempy_labs._icons as icons
from ._basic_functions import list_workspaces
from sempy._utils._log import log
from .._helper_functions import (
    _base_api,
    _is_valid_uuid,
    _build_url,
    resolve_workspace_name,
)


@log
def scan_workspaces(
    data_source_details: bool = False,
    dataset_schema: bool = False,
    dataset_expressions: bool = False,
    lineage: bool = False,
    artifact_users: bool = False,
    workspace: Optional[str | List[str] | UUID | List[UUID]] = None,
) -> dict:
    """
    Gets the scan result for the specified scan.

    This is a wrapper function for the following APIs:
        `Admin - WorkspaceInfo PostWorkspaceInfo <https://learn.microsoft.com/rest/api/power-bi/admin/workspace-info-post-workspace-info>`_.
        `Admin - WorkspaceInfo GetScanStatus <https://learn.microsoft.com/rest/api/power-bi/admin/workspace-info-get-scan-status>`_.
        `Admin - WorkspaceInfo GetScanResult <https://learn.microsoft.com/rest/api/power-bi/admin/workspace-info-get-scan-result>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    data_source_details : bool, default=False
        Whether to return dataset expressions (DAX and Mashup queries). If you set this parameter to true, you must fully enable metadata scanning in order for data to be returned. For more information, see Enable tenant settings for metadata scanning.
    dataset_schema: bool = False
        Whether to return dataset schema (tables, columns and measures). If you set this parameter to true, you must fully enable metadata scanning in order for data to be returned. For more information, see Enable tenant settings for metadata scanning.
    dataset_expressions : bool, default=False
        Whether to return data source details.
    lineage : bool, default=False
        Whether to return lineage info (upstream dataflows, tiles, data source IDs)
    artifact_users : bool, default=False
        Whether to return user details for a Power BI item (such as a report or a dashboard).
    workspace : str | List[str] | UUID | List[UUID], default=None
        The required workspace name(s) or id(s) to be scanned. It supports a limit of 100 workspaces and only IDs in GUID format.

    Returns
    -------
    dict
        A json object with the scan result.
    """

    if workspace is None:
        workspace = resolve_workspace_name()

    if isinstance(workspace, str):
        workspace = [workspace]

    if len(workspace) > 100:
        print(
            f"{icons.yellow_dot} More than 100 workspaces where provided. Truncating to the fist 100."
        )
        workspace = workspace[:100]

    workspace_list = []

    for w in workspace:
        if _is_valid_uuid(w):
            workspace_list.append(w)
        else:
            dfW = list_workspaces(workspace=w)
            workspace_list = (
                workspace_list + dfW[dfW["Name"].isin(workspace)]["Id"].tolist()
            )

    url = "/v1.0/myorg/admin/workspaces/getInfo"
    params = {}
    params["lineage"] = lineage
    params["datasourceDetails"] = data_source_details
    params["datasetSchema"] = dataset_schema
    params["datasetExpressions"] = dataset_expressions
    params["getArtifactUsers"] = artifact_users

    url = _build_url(url, params)

    payload = {"workspaces": workspace_list}

    response = _base_api(
        request=url,
        method="post",
        payload=payload,
        status_codes=202,
        client="fabric_sp",
    )

    scan_id = response.json()["id"]
    scan_status = response.json().get("status")

    while scan_status not in ["Succeeded", "Failed"]:
        time.sleep(1)
        response = _base_api(
            request=f"/v1.0/myorg/admin/workspaces/scanStatus/{scan_id}",
            client="fabric_sp",
        )
        scan_status = response.json().get("status")

    if scan_status == "Failed":
        raise FabricHTTPException(response)

    response = _base_api(
        request=f"/v1.0/myorg/admin/workspaces/scanResult/{scan_id}",
        client="fabric_sp",
    )

    return response.json()



import pandas as pd
from typing import Optional
from sempy_labs._helper_functions import (
    _base_api,
    resolve_item_id,
    resolve_workspace_id,
    resolve_workspace_name_and_id,
    resolve_item_name_and_id,
    _create_dataframe,
)
from uuid import UUID
import sempy_labs._icons as icons
from sempy._utils._log import log


@log
def get_eventstream_destination(
    eventstream: str | UUID,
    destination_id: UUID,
    workspace: Optional[str | UUID] = None,
) -> pd.DataFrame:
    """
    Returns the specified destination of the eventstream.

    This is a wrapper function for the following API: `Topology - Get Eventstream Destination <https://learn.microsoft.com/rest/api/fabric/eventstream/topology/get-eventstream-destination>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    eventstream : str | uuid.UUID
        The name or ID of the eventstream.
    destination_id : uuid.UUID
        The ID of the destination.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the details of the destination.
    """

    workspace_id = resolve_workspace_id(workspace)
    eventstream_id = resolve_item_id(
        item=eventstream, type="Eventstream", workspace=workspace_id
    )
    response = _base_api(
        request=f"/v1/workspaces/{workspace_id}/eventstreams/{eventstream_id}/destinations/{destination_id}",
        client="fabric_sp",
    )

    columns = {
        "Eventstream Destination Id": "str",
        "Eventstream Destination Name": "str",
        "Eventstream Destination Type": "str",
        "Workspace Id": "str",
        "Item Id": "str",
        "Schema": "str",
        "Delta Table": "str",
        "Input Serialization Type": "str",
        "Input Serialization Encoding": "str",
        "Input Nodes": "str",
        "Status": "str",
        "Error": "str",
    }

    df = _create_dataframe(columns)

    result = response.json()

    rows = []
    prop = result.get("properties", {})
    rows.append(
        {
            "Eventstream Destination Id": result.get("id"),
            "Eventstream Destination Name": result.get("name"),
            "Eventstream Destination Type": result.get("type"),
            "Workspace Id": prop.get("workspaceId"),
            "Item Id": prop.get("itemId"),
            "Schema": prop.get("schema"),
            "Delta Table": prop.get("deltaTable"),
            "Input Serialization Type": prop.get("inputSerialization", {}).get("type"),
            "Input Serialization Encoding": prop.get("inputSerialization", {})
            .get("properties", {})
            .get("encoding"),
            "Input Nodes": (
                ", ".join([node.get("name") for node in result.get("inputNodes", [])])
                if result.get("inputNodes")
                else None
            ),
            "Status": result.get("status"),
            "Error": result.get("error"),
        }
    )

    if rows:
        df = pd.DataFrame(rows, columns=columns.keys())

    return df


@log
def get_eventstream_destination_connection(
    eventstream: str | UUID,
    destination_id: UUID,
    workspace: Optional[str | UUID] = None,
) -> pd.DataFrame:
    """
    Returns the connection information of a specified destination of the eventstream. Only custom endpoints destinations are supported.

    This is a wrapper function for the following API: `Topology - Get Eventstream Destination Connection <https://learn.microsoft.com/rest/api/fabric/eventstream/topology/get-eventstream-destination-connection>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    eventstream : str | uuid.UUID
        The name or ID of the eventstream.
    destination_id : uuid.UUID
        The ID of the destination.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the connection details of the destination.
    """

    workspace_id = resolve_workspace_id(workspace)
    eventstream_id = resolve_item_id(
        item=eventstream, type="Eventstream", workspace=workspace_id
    )
    response = _base_api(
        request=f"/v1/workspaces/{workspace_id}/eventstreams/{eventstream_id}/destinations/{destination_id}/connection",
        client="fabric_sp",
    )

    columns = {
        "Fully Qualified Namespace": "str",
        "EventHub Name": "str",
        "Consumer Group Name": "str",
        "Primary Key": "str",
        "Secondary Key": "str",
        "Primary Connection String": "str",
        "Secondary Connection String": "str",
    }

    df = _create_dataframe(columns=columns)

    result = response.json()

    rows = []
    rows.append(
        {
            "Fully Qualified Namespace": result.get("fullyQualifiedNamespace"),
            "EventHub Name": result.get("eventHubName"),
            "Consumer Group Name": result.get("consumerGroupName"),
            "Primary Key": result.get("accessKeys", {}).get("primaryKey"),
            "Secondary Key": result.get("accessKeys", {}).get("secondaryKey"),
            "Primary Connection String": result.get("accessKeys", {}).get(
                "primaryConnectionString"
            ),
            "Secondary Connection String": result.get("accessKeys", {}).get(
                "secondaryConnectionString"
            ),
        }
    )

    if rows:
        df = pd.DataFrame(rows, columns=columns.keys())

    return df


@log
def get_eventstream_source(
    eventstream: str | UUID, source_id: UUID, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Returns the specified source of the eventstream.

    This is a wrapper function for the following API: `Topology - Get Eventstream Source <https://learn.microsoft.com/rest/api/fabric/eventstream/topology/get-eventstream-source>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    eventstream : str | uuid.UUID
        The name or ID of the eventstream.
    source_id : uuid.UUID
        The ID of the source.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the details of the source.
    """

    workspace_id = resolve_workspace_id(workspace)
    eventstream_id = resolve_item_id(
        item=eventstream, type="Eventstream", workspace=workspace_id
    )
    response = _base_api(
        request=f"/v1/workspaces/{workspace_id}/eventstreams/{eventstream_id}/sources/{source_id}",
        client="fabric_sp",
    )

    columns = {
        "Eventstream Source Id": "str",
        "Eventstream Source Name": "str",
        "Eventstream Source Type": "str",
        "Data Connection Id": "str",
        "Consumer Group Name": "str",
        "Input Serialization Type": "str",
        "Input Serialization Encoding": "str",
        "Status": "str",
        "Error": "str",
    }

    df = _create_dataframe(columns=columns)

    rows = []
    result = response.json()
    prop = result.get("properties", {})
    rows.append(
        {
            "Eventstream Source Id": result.get("id"),
            "Eventstream Source Name": result.get("name"),
            "Eventstream Source Type": result.get("type"),
            "Data Connection Id": prop.get("dataConnectionId"),
            "Consumer Group Name": prop.get("consumerGroupName"),
            "Input Serialization Type": prop.get("inputSerialization", {}).get("type"),
            "Input Serialization Encoding": prop.get("inputSerialization", {})
            .get("properties", {})
            .get("encoding"),
            "Status": result.get("status"),
            "Error": result.get("error"),
        }
    )

    if rows:
        df = pd.DataFrame(rows, columns=columns.keys())

    return df


@log
def get_eventstream_source_connection(
    eventstream: str | UUID, source_id: UUID, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Returns the connection information of specified source of the eventstream. Only custom endpoints sources are supported.

    This is a wrapper function for the following API: `Topology - Get Eventstream Source Connection <https://learn.microsoft.com/rest/api/fabric/eventstream/topology/get-eventstream-source-connection>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    eventstream : str | uuid.UUID
        The name or ID of the eventstream.
    source_id : uuid.UUID
        The ID of the source.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the connection details of the source.
    """

    workspace_id = resolve_workspace_id(workspace)
    eventstream_id = resolve_item_id(
        item=eventstream, type="Eventstream", workspace=workspace_id
    )
    response = _base_api(
        request=f"/v1/workspaces/{workspace_id}/eventstreams/{eventstream_id}/sources/{source_id}/connection",
        client="fabric_sp",
    )
    result = response.json()

    columns = {
        "Fully Qualified Namespace": "str",
        "EventHub Name": "str",
        "Consumer Group Name": "str",
        "Primary Key": "str",
        "Secondary Key": "str",
        "Primary Connection String": "str",
        "Secondary Connection String": "str",
    }
    df = _create_dataframe(columns=columns)

    rows = []
    rows.append(
        {
            "Fully Qualified Namespace": result.get("fullyQualifiedNamespace"),
            "EventHub Name": result.get("eventHubName"),
            "Consumer Group Name": result.get("consumerGroupName"),
            "Primary Key": result.get("accessKeys", {}).get("primaryKey"),
            "Secondary Key": result.get("accessKeys", {}).get("secondaryKey"),
            "Primary Connection String": result.get("accessKeys", {}).get(
                "primaryConnectionString"
            ),
            "Secondary Connection String": result.get("accessKeys", {}).get(
                "secondaryConnectionString"
            ),
        }
    )

    if rows:
        df = pd.DataFrame(rows, columns=columns.keys())

    return df


@log
def get_eventstream_topology(
    eventstream: str | UUID, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Returns the topology of the specified eventstream.

    This is a wrapper function for the following API: `Topology - Get Eventstream Topology <https://learn.microsoft.com/rest/api/fabric/eventstream/topology/get-eventstream-topology>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    eventstream : str | uuid.UUID
        The name or ID of the eventstream.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the topology of the eventstream, including sources and destinations.
    """

    workspace_id = resolve_workspace_id(workspace)
    eventstream_id = resolve_item_id(
        item=eventstream, type="Eventstream", workspace=workspace_id
    )
    response = _base_api(
        request=f"/v1/workspaces/{workspace_id}/eventstreams/{eventstream_id}/topology",
        client="fabric_sp",
    )

    columns = {
        "Eventstream Source Id": "str",
        "Eventstream Source Name": "str",
        "Eventstream Source Type": "str",
        "Data Connection Id": "str",
        "Consumer Group Name": "str",
        "Input Serialization Type": "str",
        "Input Serialization Encoding": "str",
        "Status": "str",
        "Error": "str",
        "Type": "str",
        "Region": "str",
        "Topic": "str",
        "Auto Offset Reset": "str",
        "SASL Mechanism": "str",
        "Security Protocol": "str",
        "Container Name": "str",
        "Database Name": "str",
        "Offset Policy": "str",
        "Table Name": "str",
        "Server Id": "str",
        "Port": "str",
        "Slot Name": "str",
    }
    df = _create_dataframe(columns=columns)
    rows = []
    for r in response.json().get("sources", []):
        prop = r.get("properties", {})
        rows.append(
            {
                "Eventstream Source Id": r.get("id"),
                "Eventstream Source Name": r.get("name"),
                "Eventstream Source Type": r.get("type"),
                "Data Connection Id": prop.get("dataConnectionId"),
                "Consumer Group Name": prop.get("consumerGroupName"),
                "Input Serialization Type": prop.get("inputSerialization", {}).get(
                    "type"
                ),
                "Input Serialization Encoding": prop.get("inputSerialization", {})
                .get("properties", {})
                .get("encoding"),
                "Status": r.get("status"),
                "Error": r.get("error"),
                "Type": prop.get("type"),
                "Region": prop.get("region"),
                "Topic": prop.get("topic"),
                "Auto Offset Reset": prop.get("autoOffsetReset"),
                "SASL Mechanism": prop.get("saslMechanism"),
                "Security Protocol": prop.get("securityProtocol"),
                "Container Name": prop.get("containerName"),
                "Database Name": prop.get("databaseName"),
                "Offset Policy": prop.get("offsetPolicy"),
                "Table Name": prop.get("tableName"),
                "Server Id": prop.get("serverId"),
                "Port": prop.get("port"),
                "Slot Name": prop.get("slotName"),
            }
        )

    if rows:
        df = pd.DataFrame(rows, columns=columns.keys())

    return df


@log
def resume_eventstream(eventstream: str | UUID, workspace: Optional[str | UUID] = None):
    """
    Resume running all supported sources and destinations of the eventstream.

    This is a wrapper function for the following API: `Topology - Resume Eventstream <https://learn.microsoft.com/rest/api/fabric/eventstream/topology/resume-eventstream>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    eventstream : str | uuid.UUID
        The name or ID of the eventstream.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (eventstream_name, eventstream_id) = resolve_item_name_and_id(
        item=eventstream, type="Eventstream", workspace=workspace_id
    )
    _base_api(
        request=f"/v1/workspaces/{workspace_id}/eventstreams/{eventstream_id}/resume",
        client="fabric_sp",
        method="post",
    )

    print(
        f"{icons.green_dot} The '{eventstream_name}' eventstream within the '{workspace_name}' workspace has been resumed."
    )


@log
def resume_eventstream_destination(
    eventstream: str | UUID,
    destination_id: UUID,
    workspace: Optional[str | UUID] = None,
):
    """
    Resume running the specified destination in the eventstream.

    This is a wrapper function for the following API: `Topology - Resume Eventstream Destination <https://learn.microsoft.com/rest/api/fabric/eventstream/topology/resume-eventstream-destination>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    eventstream : str | uuid.UUID
        The name or ID of the eventstream.
    destination_id : uuid.UUID
        The ID of the destination.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (eventstream_name, eventstream_id) = resolve_item_name_and_id(
        item=eventstream, type="Eventstream", workspace=workspace_id
    )
    _base_api(
        request=f"/v1/workspaces/{workspace_id}/eventstreams/{eventstream_id}/destinations/{destination_id}/resume",
        client="fabric_sp",
        method="post",
    )

    print(
        f"{icons.green_dot} The '{destination_id}' destination in the '{eventstream_name}' eventstream within the '{workspace_name}' workspace has been resumed."
    )


@log
def resume_eventstream_source(
    eventstream: str | UUID, source_id: UUID, workspace: Optional[str | UUID] = None
):
    """
    Resume running the specified source in the eventstream.

    This is a wrapper function for the following API: `Topology - Resume Eventstream Source <https://learn.microsoft.com/rest/api/fabric/eventstream/topology/resume-eventstream-source>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    eventstream : str | uuid.UUID
        The name or ID of the eventstream.
    source_id : uuid.UUID
        The ID of the source.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (eventstream_name, eventstream_id) = resolve_item_name_and_id(
        item=eventstream, type="Eventstream", workspace=workspace_id
    )
    _base_api(
        request=f"/v1/workspaces/{workspace_id}/eventstreams/{eventstream_id}/sources/{source_id}/resume",
        client="fabric_sp",
        method="post",
    )

    print(
        f"{icons.green_dot} The '{source_id}' source in the '{eventstream_name}' eventstream within the '{workspace_name}' workspace has been resumed."
    )


@log
def pause_eventstream(
    eventstream: str | UUID, workspace: Optional[str | UUID] = None
) -> dict:
    """
    Pause running all supported sources and destinations of the eventstream.

    This is a wrapper function for the following API: `Topology - Pause Eventstream <https://learn.microsoft.com/rest/api/fabric/eventstream/topology/pause-eventstream>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    eventstream : str | uuid.UUID
        The name or ID of the eventstream.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (eventstream_name, eventstream_id) = resolve_item_name_and_id(
        item=eventstream, type="Eventstream", workspace=workspace_id
    )
    _base_api(
        request=f"/v1/workspaces/{workspace_id}/eventstreams/{eventstream_id}/pause",
        client="fabric_sp",
        method="post",
    )

    print(
        f"{icons.green_dot} The '{eventstream_name}' eventstream within the '{workspace_name}' workspace has been paused."
    )


@log
def pause_eventstream_destination(
    eventstream: str | UUID,
    destination_id: UUID,
    workspace: Optional[str | UUID] = None,
):
    """
    Pause running the specified destination in the eventstream.

    This is a wrapper function for the following API: `Topology - Pause Eventstream Destination <https://learn.microsoft.com/rest/api/fabric/eventstream/topology/pause-eventstream-destination>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    eventstream : str | uuid.UUID
        The name or ID of the eventstream.
    destination_id : uuid.UUID
        The ID of the destination.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (eventstream_name, eventstream_id) = resolve_item_name_and_id(
        item=eventstream, type="Eventstream", workspace=workspace_id
    )
    _base_api(
        request=f"/v1/workspaces/{workspace_id}/eventstreams/{eventstream_id}/destinations/{destination_id}/pause",
        client="fabric_sp",
        method="post",
    )

    print(
        f"{icons.green_dot} The '{destination_id}' destination in the '{eventstream_name}' eventstream within the '{workspace_name}' workspace has been paused."
    )


@log
def pause_eventstream_source(
    eventstream: str | UUID, source_id: UUID, workspace: Optional[str | UUID] = None
):
    """
    Pause running the specified source in the eventstream.

    This is a wrapper function for the following API: `Topology - Pause Eventstream Source <https://learn.microsoft.com/rest/api/fabric/eventstream/topology/pause-eventstream-source>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    eventstream : str | uuid.UUID
        The name or ID of the eventstream.
    source_id : uuid.UUID
        The ID of the source.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (eventstream_name, eventstream_id) = resolve_item_name_and_id(
        item=eventstream, type="Eventstream", workspace=workspace_id
    )
    _base_api(
        request=f"/v1/workspaces/{workspace_id}/eventstreams/{eventstream_id}/sources/{source_id}/pause",
        client="fabric_sp",
        method="post",
    )

    print(
        f"{icons.green_dot} The '{source_id}' source in the '{eventstream_name}' eventstream within the '{workspace_name}' workspace has been paused."
    )



import requests
import pandas as pd
from sempy.fabric.exceptions import FabricHTTPException
from sempy._utils._log import log
import sempy_labs._icons as icons
from typing import Optional
from uuid import UUID
from sempy_labs._kql_databases import _resolve_cluster_uri
from sempy_labs._helper_functions import resolve_item_id


@log
def query_kusto(
    query: str,
    kql_database: str | UUID,
    workspace: Optional[str | UUID] = None,
    language: str = "kql",
) -> pd.DataFrame:
    """
    Runs a KQL query against a KQL database.

    Parameters
    ----------
    query : str
        The query (supports KQL or SQL - make sure to specify the language parameter accordingly).
    kql_database : str | uuid.UUID
        The KQL database name or ID.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    language : str, default="kql"
        The language of the query. Currently "kql' and "sql" are supported.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the result of the KQL query.
    """

    import notebookutils

    language = language.lower()
    if language not in ["kql", "sql"]:
        raise ValueError(
            f"{icons._red_dot} Invalid language '{language}'. Only 'kql' and 'sql' are supported."
        )

    cluster_uri = _resolve_cluster_uri(kql_database=kql_database, workspace=workspace)
    token = notebookutils.credentials.getToken(cluster_uri)

    headers = {
        "Authorization": f"Bearer {token}",
        "Content-Type": "application/json",
        "Accept": "application/json",
    }

    kql_database_id = resolve_item_id(
        item=kql_database, type="KQLDatabase", workspace=workspace
    )
    payload = {"db": kql_database_id, "csl": query}
    if language == "sql":
        payload["properties"] = {"Options": {"query_language": "sql"}}

    response = requests.post(
        f"{cluster_uri}/v1/rest/query",
        headers=headers,
        json=payload,
    )

    if response.status_code != 200:
        raise FabricHTTPException(response)

    results = response.json()
    columns_info = results["Tables"][0]["Columns"]
    rows = results["Tables"][0]["Rows"]

    df = pd.DataFrame(rows, columns=[col["ColumnName"] for col in columns_info])

    return df
    # for col_info in columns_info:
    #    col_name = col_info["ColumnName"]
    #    data_type = col_info["DataType"]

    #    try:
    #        if data_type == "DateTime":
    #            df[col_name] = pd.to_datetime(df[col_name])
    #        elif data_type in ["Int64", "Int32", "Long"]:
    #            df[col_name] = (
    #                pd.to_numeric(df[col_name], errors="coerce")
    #                .fillna(0)
    #                .astype("int64")
    #            )
    #        elif data_type == "Real" or data_type == "Double":
    #            df[col_name] = pd.to_numeric(df[col_name], errors="coerce")
    #        else:
    #            # Convert any other type to string, change as needed
    #            df[col_name] = df[col_name].astype(str)
    #    except Exception as e:
    #        print(
    #            f"{icons.yellow_dot} Could not convert column {col_name} to {data_type}, defaulting to string: {str(e)}"
    #        )
    #        df[col_name] = df[col_name].astype(str)

    return df


@log
def query_workspace_monitoring(
    query: str, workspace: Optional[str | UUID] = None, language: str = "kql"
) -> pd.DataFrame:
    """
    Runs a query against the Fabric workspace monitoring database. Workspace monitoring must be enabled on the workspace to use this function.

    Parameters
    ----------
    query : str
        The query (supports KQL or SQL - make sure to specify the language parameter accordingly).
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    language : str, default="kql"
        The language of the query. Currently "kql' and "sql" are supported.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the result of the query.
    """

    return query_kusto(
        query=query,
        kql_database="Monitoring KQL database",
        workspace=workspace,
        language=language,
    )



import sempy.fabric as fabric
import numpy as np
import pandas as pd
from typing import Optional, List, Union, Tuple
from uuid import UUID
import sempy_labs._icons as icons
from sempy._utils._log import log
from .._helper_functions import (
    retry,
    _convert_data_type,
    resolve_dataset_name_and_id,
    resolve_workspace_name_and_id,
)


@log
def check_fallback_reason(
    dataset: str | UUID, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Shows the reason a table in a Direct Lake semantic model would fallback to DirectQuery.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        The tables in the semantic model and their fallback reason.
    """
    from sempy_labs.tom import connect_semantic_model

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(
        dataset, workspace=workspace_id
    )

    with connect_semantic_model(
        dataset=dataset_id, workspace=workspace_id, readonly=True
    ) as tom:
        if not tom.is_direct_lake():
            raise ValueError(
                f"{icons.red_dot} The '{dataset_name}' semantic model is not in Direct Lake. This function is only applicable to Direct Lake semantic models."
            )

    df = fabric.evaluate_dax(
        dataset=dataset_id,
        workspace=workspace_id,
        dax_string="""
            SELECT [TableName] AS [Table Name],[FallbackReason] AS [FallbackReasonID]
            FROM $SYSTEM.TMSCHEMA_DELTA_TABLE_METADATA_STORAGES
            """,
    )

    value_mapping = {
        0: "No reason for fallback",
        1: "This table is not framed",
        2: "This object is a view in the lakehouse",
        3: "The table does not exist in the lakehouse",
        4: "Transient error",
        5: "Using OLS will result in fallback to DQ",
        6: "Using RLS will result in fallback to DQ",
    }

    # Create a new column based on the mapping
    df["Fallback Reason Detail"] = np.vectorize(value_mapping.get)(
        df["FallbackReasonID"]
    )

    return df


@log
def generate_direct_lake_semantic_model(
    dataset: str,
    lakehouse_tables: Union[str, List[str]],
    workspace: Optional[str | UUID] = None,
    lakehouse: Optional[str] = None,
    lakehouse_workspace: Optional[str | UUID] = None,
    schema: str = "dbo",
    overwrite: bool = False,
    refresh: bool = True,
):
    """
    Dynamically generates a Direct Lake semantic model based on tables in a Fabric lakehouse.

    Parameters
    ----------
    dataset : str
        Name of the semantic model to be created.
    lakehouse_tables : str | List[str]
        The table(s) within the Fabric lakehouse to add to the semantic model. All columns from these tables will be added to the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID in which the semantic model will reside.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    lakehouse : str, default=None
        The lakehouse which stores the delta tables which will feed the Direct Lake semantic model.
        Defaults to None which resolves to the attached lakehouse.
    lakehouse_workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID in which the lakehouse resides.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    schema : str, default="dbo"
        The schema used for the lakehouse.
    overwrite : bool, default=False
        If set to True, overwrites the existing semantic model if it already exists.
    refresh: bool, default=True
        If True, refreshes the newly created semantic model after it is created.
    """

    from sempy_labs.lakehouse import get_lakehouse_tables, get_lakehouse_columns
    from sempy_labs.directlake._generate_shared_expression import (
        generate_shared_expression,
    )
    from sempy_labs.tom import connect_semantic_model
    from sempy_labs._generate_semantic_model import create_blank_semantic_model
    from sempy_labs._refresh_semantic_model import refresh_semantic_model

    if isinstance(lakehouse_tables, str):
        lakehouse_tables = [lakehouse_tables]

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    if lakehouse_workspace is None:
        lakehouse_workspace = workspace

    dfLT = get_lakehouse_tables(lakehouse=lakehouse, workspace=lakehouse_workspace)

    icons.sll_tags.append("GenerateDLModel")

    # Validate lakehouse tables
    for t in lakehouse_tables:
        if t not in dfLT["Table Name"].values:
            raise ValueError(
                f"{icons.red_dot} The '{t}' table does not exist as a delta table in the '{lakehouse}' within the '{workspace_name}' workspace."
            )

    dfLC = get_lakehouse_columns(lakehouse=lakehouse, workspace=lakehouse_workspace)
    expr = generate_shared_expression(
        item_name=lakehouse, item_type="Lakehouse", workspace=lakehouse_workspace
    )
    dfD = fabric.list_datasets(workspace=workspace_id, mode="rest")
    dfD_filt = dfD[dfD["Dataset Name"] == dataset]

    if len(dfD_filt) > 0 and not overwrite:
        raise ValueError(
            f"{icons.red_dot} The '{dataset}' semantic model within the '{workspace_name}' workspace already exists. Overwrite is set to False so the new semantic model has not been created."
        )

    create_blank_semantic_model(
        dataset=dataset, workspace=workspace_id, overwrite=overwrite
    )

    @retry(
        sleep_time=1,
        timeout_error_message=f"{icons.red_dot} Function timed out after 1 minute",
    )
    def dyn_connect():
        with connect_semantic_model(
            dataset=dataset, readonly=True, workspace=workspace_id
        ) as tom:

            tom.model

    dyn_connect()

    expression_name = "DatabaseQuery"
    with connect_semantic_model(
        dataset=dataset, workspace=workspace_id, readonly=False
    ) as tom:
        if not any(e.Name == expression_name for e in tom.model.Expressions):
            tom.add_expression(name=expression_name, expression=expr)

        for t in lakehouse_tables:
            tom.add_table(name=t)
            tom.add_entity_partition(table_name=t, entity_name=t, schema_name=schema)
            dfLC_filt = dfLC[dfLC["Table Name"] == t]
            for i, r in dfLC_filt.iterrows():
                lakeCName = r["Column Name"]
                dType = r["Data Type"]
                dt = _convert_data_type(dType)
                tom.add_data_column(
                    table_name=t,
                    column_name=lakeCName,
                    source_column=lakeCName,
                    data_type=dt,
                )

    if refresh:
        refresh_semantic_model(dataset=dataset, workspace=workspace_id)


@log
def get_direct_lake_source(
    dataset: str | UUID, workspace: Optional[str | UUID] = None
) -> Tuple[str, str, UUID, UUID]:
    """
    Obtains the source information for a direct lake semantic model (if the source is located in the same workspace as the semantic model).

    Parameters
    ----------
    dataset : str | uuid.UUID
        The name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    Tuple[str, str, UUID, UUID]
        If the source of the direct lake semantic model is a lakehouse this will return: 'Lakehouse', Lakehouse Name, SQL Endpoint Id, Workspace Id
        If the source of the direct lake semantic model is a warehouse this will return: 'Warehouse', Warehouse Name, Warehouse Id, Workspace Id
        If the semantic model is not a Direct Lake semantic model, it will return None, None, None.
    """

    from sempy_labs._helper_functions import get_direct_lake_sql_endpoint

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    sql_endpoint_id = get_direct_lake_sql_endpoint(dataset=dataset, workspace=workspace)
    dfI = fabric.list_items(workspace=workspace)
    dfI_filt = dfI[
        (dfI["Id"] == sql_endpoint_id)
        & (dfI["Type"].isin(["SQLEndpoint", "Warehouse"]))
    ]

    artifact_type, artifact_name, artifact_id = None, None, None

    if not dfI_filt.empty:
        artifact_name = dfI_filt["Display Name"].iloc[0]
        artifact_id = dfI[
            (dfI["Display Name"] == artifact_name)
            & (dfI["Type"].isin(["Lakehouse", "Warehouse"]))
        ]["Id"].iloc[0]
        artifact_type = dfI[
            (dfI["Display Name"] == artifact_name)
            & (dfI["Type"].isin(["Lakehouse", "Warehouse"]))
        ]["Type"].iloc[0]

    return artifact_type, artifact_name, artifact_id, workspace_id

    # payload = {
    #   "artifacts": [
    #       {
    #           "objectId": dataset_id,
    #           "type": "dataset",
    #       }
    #   ]
    # }

    # response = _base_api(request="metadata/relations/upstream?apiVersion=3", payload=payload, method="post")

    # artifacts = response.json().get("artifacts", [])
    # sql_id, sql_object_name, sql_workspace_id, artifact_type = None, None, None, None

    # for artifact in artifacts:
    #    object_type = artifact.get("typeName")
    #    display_name = artifact.get("displayName")
    #    if object_type in ["Datawarehouse", "Lakewarehouse"]:
    #        artifact_type = (
    #            "Warehouse" if object_type == "Datawarehouse" else "Lakehouse"
    #        )
    #        sql_id = artifact.get("objectId")
    #        sql_workspace_id = artifact.get("workspace", {}).get("objectId")
    #        sql_object_name = display_name
    #        break

    # return artifact_type, sql_object_name, sql_id, sql_workspace_id



import requests
import sempy_labs._icons as icons


vis_type_mapping = {
    "barChart": "Bar chart",
    "columnChart": "Column chart",
    "clusteredBarChart": "Clustered bar chart",
    "clusteredColumnChart": "Clustered column chart",
    "hundredPercentStackedBarChart": "100% Stacked bar chart",
    "hundredPercentStackedColumnChart": "100% Stacked column chart",
    "lineChart": "Line chart",
    "areaChart": "Area chart",
    "stackedAreaChart": "Stacked area chart",
    "lineStackedColumnComboChart": "Line and stacked column chart",
    "lineClusteredColumnComboChart": "Line and clustered column chart",
    "ribbonChart": "Ribbon chart",
    "waterfallChart": "Waterfall chart",
    "funnel": "Funnel chart",
    "scatterChart": "Scatter chart",
    "pieChart": "Pie chart",
    "donutChart": "Donut chart",
    "treemap": "Treemap",
    "map": "Map",
    "filledMap": "Filled map",
    "shapeMap": "Shape map",
    "azureMap": "Azure map",
    "gauge": "Gauge",
    "card": "Card",
    "multiRowCard": "Multi-row card",
    "kpi": "KPI",
    "slicer": "Slicer",
    "tableEx": "Table",
    "pivotTable": "Matrix",
    "scriptVisual": "R script visual",
    "pythonVisual": "Python visual",
    "keyDriversVisual": "Key influencers",
    "decompositionTreeVisual": "Decomposition tree",
    "qnaVisual": "Q&A",
    "aiNarratives": "Narrative",
    "scorecard": "Goals (Preview)",
    "rdlVisual": "Paginated report",
    "cardVisual": "Card (new)",
    "actionButton": "Button",
    "bookmarkNavigator": "Bookmark navigator",
    "image": "Image",
    "textbox": "Textbox",
    "pageNavigator": "Page navigator",
    "shape": "Shape",
    "Group": "Group",
    "listSlicer": "List Slicer",
    "advancedSlicerVisual": "Button Slicer",
    "FlowVisual_C29F1DCC_81F5_4973_94AD_0517D44CC06A": "Power Automate for Power BI",
}


def generate_visual_file_path(page_file_path: str, visual_id: str) -> str:

    return page_file_path.split("/page.json")[0] + f"/visuals/{visual_id}.json"


def resolve_visual_type(visual_type: str) -> str:
    vt_lower = visual_type.lower()

    vis_map_lower = {k.lower(): v for k, v in vis_type_mapping.items()}
    flipped_lower = {v.lower(): k for k, v in vis_type_mapping.items()}

    if vt_lower in vis_map_lower:
        resolved = vis_map_lower.get(vt_lower)
    elif vt_lower in flipped_lower:
        resolved = flipped_lower.get(vt_lower)
    else:
        raise ValueError(f"{icons.red_dot} Unknown visual type: {visual_type}")

    return resolved


page_type_mapping = {
    (320, 240): "Tooltip",
    (816, 1056): "Letter",
    (960, 720): "4:3",
    (1280, 720): "16:9",
}

page_types = ["Tooltip", "Letter", "4:3", "16:9"]


def populate_custom_visual_display_names():

    url = "https://catalogapi.azure.com/offers?api-version=2018-08-01-beta&storefront=appsource&$filter=offerType+eq+%27PowerBIVisuals%27"

    def fetch_all_pages(start_url):
        combined_json = {}
        current_url = start_url

        while current_url:
            # Send GET request to the current page URL
            response = requests.get(current_url)

            if response.status_code == 200:
                data = response.json()
                # Merge the current page JSON into the combined JSON
                for key, value in data.items():
                    if key not in combined_json:
                        combined_json[key] = value
                    else:
                        # If the key already exists and is a list, extend it
                        if isinstance(value, list):
                            combined_json[key].extend(value)
                        # For other types (non-lists), update the value
                        else:
                            combined_json[key] = value

                # Get the next page link if it exists
                current_url = data.get("nextPageLink")
            else:
                print(f"Error fetching page: {response.status_code}")
                break

        return combined_json

    cvJson = fetch_all_pages(url)

    for i in cvJson.get("items", []):
        vizId = i.get("powerBIVisualId")
        displayName = i.get("displayName")
        vis_type_mapping[vizId] = displayName


def find_entity_property_pairs(data, result=None, keys_path=None):

    if result is None:
        result = {}
    if keys_path is None:
        keys_path = []

    if isinstance(data, dict):
        expression = data.get("Expression", {})
        source_ref = (
            expression.get("SourceRef", {}) if isinstance(expression, dict) else {}
        )

        if (
            isinstance(source_ref, dict)
            and "Entity" in source_ref
            and "Property" in data
        ):
            entity = source_ref.get("Entity", "")
            property_value = data.get("Property", "")

            object_type = (
                keys_path[-1].replace("HierarchyLevel", "Hierarchy")
                if keys_path
                else "Unknown"
            )
            result[property_value] = (entity, object_type)
            if keys_path:
                keys_path.pop()

        # Recursively search the rest of the dictionary
        for key, value in data.items():
            keys_path.append(key)
            find_entity_property_pairs(value, result, keys_path)

    elif isinstance(data, list):
        for item in data:
            find_entity_property_pairs(item, result, keys_path)

    return result


def _get_agg_type_mapping() -> dict:
    """
    This function extracts a mapping dictionary like this:
    {
        "0": "Sum",
        "1": "Average",
        "2": "Distinct count",
    }
    """

    schema_url = "https://developer.microsoft.com/json-schemas/fabric/item/report/definition/semanticQuery/1.2.0/schema.json"
    response = requests.get(schema_url)
    schema = response.json()
    aggtypes_schema = schema.get("definitions", {}).get("QueryAggregateFunction", {})

    agg_type_map = {}
    agg_type_map = {
        a.get("const"): a.get("description")
        for a in aggtypes_schema.get("anyOf", [])
        if "const" in a and "description" in a
    }
    agg_type_map["-1"] = "Unknown"

    return agg_type_map


def _get_expression(expr_json, agg_type_map):

    expr_type = list(expr_json.keys())[0]
    if expr_type == "Literal":
        expr = expr_json.get("Literal", {}).get("Value")[1:-1]
    elif expr_type == "Aggregation":
        entity = (
            expr_json.get("Aggregation", {})
            .get("Expression", {})
            .get("Column", {})
            .get("Expression", {})
            .get("SourceRef", {})
            .get("Entity", "Unknown")
        )
        column = (
            expr_json.get("Aggregation", {})
            .get("Expression", {})
            .get("Column", {})
            .get("Property", "Unknown")
        )
        function_id = expr_json.get("Aggregation", {}).get("Function", "-1")
        function = agg_type_map.get(function_id)
        expr = f"{function}('{entity}'[{column}])"
    elif expr_type == "Measure":
        measure = expr_json.get("Measure", {}).get("Property", "Unknown")
        expr = f"[{measure}]"
    else:
        expr = "Unknown"

    return expr



import sempy.fabric as fabric
from .._helper_functions import (
    resolve_lakehouse_id,
)
from typing import Optional, Tuple
from uuid import UUID
import sempy_labs._icons as icons
from sempy._utils._log import log


@log
def get_direct_lake_lakehouse(
    dataset: str | UUID,
    workspace: Optional[str | UUID] = None,
    lakehouse: Optional[str] = None,
    lakehouse_workspace: Optional[str | UUID] = None,
) -> Tuple[str, UUID]:
    """
    Identifies the lakehouse used by a Direct Lake semantic model.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    lakehouse : str, default=None
        The Fabric lakehouse used by the Direct Lake semantic model.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    lakehouse_workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID used by the lakehouse.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    Tuple[str, uuid.UUID]
        The lakehouse name and lakehouse ID.
    """

    from sempy_labs.directlake._dl_helper import get_direct_lake_source

    artifact_type, artifact_name, artifact_id, workspace_id = get_direct_lake_source(
        dataset=dataset, workspace=workspace
    )

    if artifact_type in ["Lakehouse", "Warehouse"]:
        return artifact_name, artifact_id
    else:
        fabric.refresh_tom_cache(workspace=workspace)
        dfP = fabric.list_partitions(dataset=dataset, workspace=workspace)
        dfP_filt = dfP[dfP["Mode"] == "DirectLake"]
        if dfP_filt.empty:
            raise ValueError(
                f"{icons.red_dot} The '{dataset}' semantic model within the '{workspace}' workspace is not in Direct Lake mode."
            )
        lakehouse_id = resolve_lakehouse_id(
            lakehouse=lakehouse, workspace=lakehouse_workspace
        )
        return lakehouse, lakehouse_id



import pandas as pd
from typing import Optional
from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    _update_dataframe_datatypes,
    _base_api,
    resolve_item_id,
    _create_dataframe,
    delete_item,
    create_item,
    _get_item_definition,
    resolve_workspace_id,
)
import sempy_labs._icons as icons
import base64
from uuid import UUID
from sempy._utils._log import log


@log
def list_mirrored_databases(workspace: Optional[str | UUID] = None) -> pd.DataFrame:
    """
    Shows the mirrored databases within a workspace.

    This is a wrapper function for the following API: `Items - List Mirrored Databases <https://learn.microsoft.com/rest/api/fabric/mirroredwarehouse/items/list-mirrored-databases>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the mirrored databases within a workspace.
    """

    columns = {
        "Mirrored Database Name": "string",
        "Mirrored Database Id": "string",
        "Description": "string",
        "OneLake Tables Path": "string",
        "SQL Endpoint Connection String": "string",
        "SQL Endpoint Id": "string",
        "Provisioning Status": "string",
        "Default Schema": "string",
    }
    df = _create_dataframe(columns=columns)

    workspace_id = resolve_workspace_id(workspace)
    responses = _base_api(
        request=f"/v1/workspaces/{workspace_id}/mirroredDatabases",
        uses_pagination=True,
        client="fabric_sp",
    )

    rows = []
    for r in responses:
        for v in r.get("value", []):
            prop = v.get("properties", {})
            sql = prop.get("sqlEndpointProperties", {})
            rows.append(
                {
                    "Mirrored Database Name": v.get("displayName"),
                    "Mirrored Database Id": v.get("id"),
                    "Description": v.get("description"),
                    "OneLake Tables Path": prop.get("oneLakeTablesPath"),
                    "SQL Endpoint Connection String": sql.get("connectionString"),
                    "SQL Endpoint Id": sql.get("id"),
                    "Provisioning Status": sql.get("provisioningStatus"),
                    "Default Schema": prop.get("defaultSchema"),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df


@log
def create_mirrored_database(
    name: str, description: Optional[str] = None, workspace: Optional[str | UUID] = None
):
    """
    Creates a Fabric mirrored database.

    This is a wrapper function for the following API: `Items - Create Mirrored Database <https://learn.microsoft.com/rest/api/fabric/mirroreddatabase/items/create-mirrored-database>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    name: str
        Name of the mirrored database.
    description : str, default=None
        A description of the mirrored database.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    create_item(
        name=name, description=description, type="MirroredDatabase", workspace=workspace
    )


@log
def delete_mirrored_database(
    mirrored_database: str, workspace: Optional[str | UUID] = None
):
    """
    Deletes a mirrored database.

    This is a wrapper function for the following API: `Items - Delete Mirrored Database <https://learn.microsoft.com/rest/api/fabric/mirroreddatabase/items/delete-mirrored-database>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    mirrored_database: str
        Name of the mirrored database.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    delete_item(item=mirrored_database, type="MirroredDatabase", workspace=workspace)


@log
def get_mirroring_status(
    mirrored_database: str | UUID, workspace: Optional[str | UUID] = None
) -> str:
    """
    Get the status of the mirrored database.

    This is a wrapper function for the following API: `Mirroring - Get Mirroring Status <https://learn.microsoft.com/rest/api/fabric/mirroreddatabase/mirroring/get-mirroring-status>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    mirrored_database: str | uuid.UUID
        Name or ID of the mirrored database.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    str
        The status of a mirrored database.
    """

    workspace_id = resolve_workspace_id(workspace)
    item_id = resolve_item_id(
        item=mirrored_database, type="MirroredDatabase", workspace=workspace
    )
    response = _base_api(
        request=f"/v1/workspaces/{workspace_id}/mirroredDatabases/{item_id}/getMirroringStatus",
        status_codes=200,
        client="fabric_sp",
    )

    return response.json().get("status", {})


@log
def get_tables_mirroring_status(
    mirrored_database: str | UUID, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Gets the mirroring status of the tables.

    This is a wrapper function for the following API: `Mirroring - Get Tables Mirroring Status <https://learn.microsoft.com/rest/api/fabric/mirroreddatabase/mirroring/get-tables-mirroring-status>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    mirrored_database: str | uuid.UUID
        Name or ID of the mirrored database.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the mirroring status of the tables.
    """

    workspace_id = resolve_workspace_id(workspace)
    item_id = resolve_item_id(
        item=mirrored_database, type="MirroredDatabase", workspace=workspace
    )
    responses = _base_api(
        request=f"/v1/workspaces/{workspace_id}/mirroredDatabases/{item_id}/getTablesMirroringStatus",
        method="post",
        status_codes=200,
        uses_pagination=True,
        client="fabric_sp",
    )

    columns = {
        "Source Schema Name": "string",
        "Source Table Name": "string",
        "Status": "string",
        "Processed Bytes": "int",
        "Processed Rows": "int",
        "Last Sync Date": "datetime",
    }
    df = _create_dataframe(columns=columns)

    rows = []
    for r in responses:
        for v in r.get("data", []):
            m = v.get("metrics", {})
            rows.append(
                {
                    "Source Schema Name": v.get("sourceSchemaName"),
                    "Source Table Name": v.get("sourceTableName"),
                    "Status": v.get("status"),
                    "Processed Bytes": m.get("processedBytes"),
                    "Processed Rows": m.get("processedRows"),
                    "Last Sync Date": m.get("lastSyncDateTime"),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))
        _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


@log
def start_mirroring(
    mirrored_database: str | UUID, workspace: Optional[str | UUID] = None
):
    """
    Starts the mirroring for a database.

    This is a wrapper function for the following API: `Mirroring - Start Mirroring <https://learn.microsoft.com/rest/api/fabric/mirroreddatabase/mirroring/start-mirroring>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    mirrored_database: str | uuid.UUID
        Name or ID of the mirrored database.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    item_id = resolve_item_id(
        item=mirrored_database, type="MirroredDatabase", workspace=workspace
    )
    _base_api(
        request=f"/v1/workspaces/{workspace_id}/mirroredDatabases/{item_id}/startMirroring",
        method="post",
        status_codes=200,
        client="fabric_sp",
    )

    print(
        f"{icons.green_dot} Mirroring has started for the '{mirrored_database}' database within the '{workspace_name}' workspace."
    )


@log
def stop_mirroring(
    mirrored_database: str | UUID, workspace: Optional[str | UUID] = None
):
    """
    Stops the mirroring for a database.

    This is a wrapper function for the following API: `Mirroring - Stop Mirroring <https://learn.microsoft.com/rest/api/fabric/mirroreddatabase/mirroring/stop-mirroring>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    mirrored_database: str | uuid.UUID
        Name or ID of the mirrored database.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    item_id = resolve_item_id(
        item=mirrored_database, type="MirroredDatabase", workspace=workspace
    )
    _base_api(
        request=f"/v1/workspaces/{workspace_id}/mirroredDatabases/{item_id}/stopMirroring",
        method="post",
        status_codes=200,
        client="fabric_sp",
    )

    print(
        f"{icons.green_dot} Mirroring has stopped for the '{mirrored_database}' database within the '{workspace_name}' workspace."
    )


@log
def get_mirrored_database_definition(
    mirrored_database: str | UUID,
    workspace: Optional[str | UUID] = None,
    decode: bool = True,
) -> dict:
    """
    Obtains the mirrored database definition.

    This is a wrapper function for the following API: `Items - Get Mirrored Database Definition <https://learn.microsoft.com/rest/api/fabric/mirroreddatabase/items/get-mirrored-database-definition>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    mirrored_database : str | uuid.UUID
        The name or ID of the mirrored database.
    workspace : str | uuid.UUID, default=None
        The name or ID of the workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    decode : bool, default=True
        If True, decodes the mirrored database definition file into .json format.
        If False, obtains the mirrored database definition file in base64 format.

    Returns
    -------
    dict
        The mirrored database definition.
    """

    return _get_item_definition(
        item=mirrored_database,
        type="MirroredDatabase",
        workspace=workspace,
        return_dataframe=False,
        decode=decode,
    )


@log
def update_mirrored_database_definition(
    mirrored_database: str | UUID,
    mirrored_database_content: dict,
    workspace: Optional[str | UUID] = None,
):
    """
    Updates an existing notebook with a new definition.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    mirrored_database : str | uuid.UUID
        The name or ID of the mirrored database to be updated.
    mirrored_database_content : dict
        The mirrored database definition (not in Base64 format).
    workspace : str | uuid.UUID, default=None
        The name or ID of the workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    item_id = resolve_item_id(
        item=mirrored_database, type="MirroredDatabase", workspace=workspace
    )
    payload = base64.b64encode(mirrored_database_content)

    request_body = {
        "displayName": mirrored_database,
        "definition": {
            "format": "ipynb",
            "parts": [
                {
                    "path": "mirroredDatabase.json",
                    "payload": payload,
                    "payloadType": "InlineBase64",
                }
            ],
        },
    }

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/mirroredDatabases/{item_id}/updateDefinition",
        method="post",
        json=request_body,
        status_codes=None,
        lro_return_status_code=True,
        client="fabric_sp",
    )

    print(
        f"{icons.green_dot} The '{mirrored_database}' mirrored database was updated within the '{workspace_name}' workspace."
    )



import sempy.fabric as fabric
import time
from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    _get_partition_map,
    _process_and_display_chart,
    resolve_dataset_name_and_id,
    _update_dataframe_datatypes,
    _base_api,
)
from typing import Any, List, Optional, Union
from sempy._utils._log import log
import sempy_labs._icons as icons
import pandas as pd
import warnings
import ipywidgets as widgets
import json
from uuid import UUID


@log
def refresh_semantic_model(
    dataset: str | UUID,
    tables: Optional[Union[str, List[str]]] = None,
    partitions: Optional[Union[str, List[str]]] = None,
    refresh_type: str = "full",
    retry_count: int = 0,
    apply_refresh_policy: bool = True,
    max_parallelism: int = 10,
    workspace: Optional[str | UUID] = None,
    visualize: bool = False,
    commit_mode: str = "transactional",
) -> pd.DataFrame | None:
    """
    Refreshes a semantic model.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    tables : str, List[str], default=None
        A string or a list of tables to refresh.
    partitions: str, List[str], default=None
        A string or a list of partitions to refresh. Partitions must be formatted as such: 'Table Name'[Partition Name].
    refresh_type : str, default="full"
        The type of processing to perform. Types align with the TMSL refresh command types: full, clearValues, calculate, dataOnly, automatic, and defragment. The add type isn't supported. Defaults to "full".
    retry_count : int, default=0
        Number of times the operation retries before failing.
    apply_refresh_policy : bool, default=True
        If an incremental refresh policy is defined, determines whether to apply the policy. Modes are true or false. If the policy isn't applied, the full process leaves partition definitions unchanged, and fully refreshes all partitions in the table. If commitMode is transactional, applyRefreshPolicy can be true or false. If commitMode is partialBatch, applyRefreshPolicy of true isn't supported, and applyRefreshPolicy must be set to false.
    max_parallelism : int, default=10
        Determines the maximum number of threads that can run the processing commands in parallel.
        This value aligns with the MaxParallelism property that can be set in the TMSL Sequence command or by using other methods.
        Defaults to 10.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    visualize : bool, default=False
        If True, displays a Gantt chart showing the refresh statistics for each table/partition.
    commit_mode : str, default="transactional"
        Determines whether to commit objects in batches or only when complete. Modes are "transactional" and "partialBatch". Defaults to "transactional".

    Returns
    -------
    pandas.DataFrame | None
        If 'visualize' is set to True, returns a pandas dataframe showing the SSAS trace output used to generate the visualization.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    if isinstance(tables, str):
        tables = [tables]
    if isinstance(partitions, str):
        partitions = [partitions]

    objects: List[Any] = []

    if tables is not None:
        objects = objects + [{"table": table} for table in tables]
    if partitions is not None:

        def extract_names(partition):
            parts = partition.split("[")
            table_name = parts[0].strip("'")
            partition_name = parts[1].strip("]")
            return {"table": table_name, "partition": partition_name}

        objects = objects + [extract_names(partition) for partition in partitions]

    refresh_type = refresh_type.lower()
    for prefix, mapped_value in icons.refresh_type_mapping.items():
        if refresh_type.startswith(prefix):
            refresh_type = mapped_value
            break

    valid_refresh_types = list(icons.refresh_type_mapping.values())
    if refresh_type not in valid_refresh_types:
        raise ValueError(
            f"{icons.red_dot} Invalid refresh type. Refresh type must be one of these values: {valid_refresh_types}."
        )

    def refresh_and_trace_dataset(
        dataset,
        workspace,
        refresh_type,
        retry_count,
        apply_refresh_policy,
        max_parallelism,
        objects,
        visualize,
        commit_mode,
    ):
        # Ignore specific warnings
        warnings.filterwarnings(
            "ignore",
            message="No trace logs have been recorded. Try starting the trace with a larger 'delay'",
        )

        def extract_failure_error():
            error_messages = []
            combined_messages = ""
            final_message = f"{icons.red_dot} The refresh of the '{dataset_name}' semantic model within the '{workspace_name}' workspace has failed."
            for _, r in fabric.get_refresh_execution_details(
                refresh_request_id=request_id,
                dataset=dataset_id,
                workspace=workspace_id,
            ).messages.iterrows():
                error_messages.append(f"{r['Type']}: {r['Message']}")

            if error_messages:
                combined_messages = "\n".join(error_messages)
            final_message += f"'\n' {combined_messages}"

            return final_message

        # Function to perform dataset refresh
        def refresh_dataset():
            return fabric.refresh_dataset(
                dataset=dataset_id,
                workspace=workspace_id,
                refresh_type=refresh_type,
                retry_count=retry_count,
                apply_refresh_policy=apply_refresh_policy,
                max_parallelism=max_parallelism,
                commit_mode=commit_mode,
                objects=objects if objects else None,
            )

        def check_refresh_status(request_id):
            request_details = fabric.get_refresh_execution_details(
                dataset=dataset_id,
                refresh_request_id=request_id,
                workspace=workspace_id,
            )
            return request_details.status

        def display_trace_logs(trace, partition_map, widget, title, stop=False):
            if stop:
                df = trace.stop()
            else:
                df = trace.get_trace_logs()
            if not df.empty:
                df = df[
                    df["Event Subclass"].isin(["ExecuteSql", "Process"])
                ].reset_index(drop=True)
                df = pd.merge(
                    df,
                    partition_map[
                        ["PartitionID", "Object Name", "TableName", "PartitionName"]
                    ],
                    left_on="Object ID",
                    right_on="PartitionID",
                    how="left",
                )
                if not df.empty:
                    _process_and_display_chart(df, title=title, widget=widget)
                if stop:
                    df.drop(["Object Name", "PartitionID"], axis=1, inplace=True)
                    df.rename(columns={"TableName": "Table Name"}, inplace=True)
                    df.rename(columns={"PartitionName": "Partition Name"}, inplace=True)
                    return df

        # Start the refresh process
        if not visualize:
            request_id = refresh_dataset()
        print(
            f"{icons.in_progress} Refresh of the '{dataset_name}' semantic model within the '{workspace_name}' workspace is in progress..."
        )

        # Monitor refresh progress and handle tracing if visualize is enabled
        if visualize:
            partition_map = _get_partition_map(dataset, workspace)
            widget = widgets.Output()

            with fabric.create_trace_connection(
                dataset=dataset_id, workspace=workspace_id
            ) as trace_connection:
                with trace_connection.create_trace(icons.refresh_event_schema) as trace:
                    trace.start()
                    request_id = refresh_dataset()

                    while True:
                        status = check_refresh_status(request_id)
                        # Check if the refresh has completed
                        if status == "Completed":
                            break
                        elif status == "Failed":
                            raise ValueError(extract_failure_error())
                        elif status == "Cancelled":
                            print(
                                f"{icons.yellow_dot} The refresh of the '{dataset_name}' semantic model within the '{workspace_name}' workspace has been cancelled."
                            )
                            return

                        # Capture and display logs in real-time
                        display_trace_logs(
                            trace,
                            partition_map,
                            widget,
                            title="Refresh in progress...",
                        )

                        time.sleep(3)  # Wait before the next check

                    # Final log display after completion
                    time.sleep(5)

                    # Stop trace and display final chart
                    final_df = display_trace_logs(
                        trace,
                        partition_map,
                        widget,
                        title="Refresh Completed",
                        stop=True,
                    )

                    print(
                        f"{icons.green_dot} Refresh '{refresh_type}' of the '{dataset_name}' semantic model within the '{workspace_name}' workspace is complete."
                    )
                    return final_df

        # For non-visualize case, only check refresh status
        else:
            while True:
                status = check_refresh_status(request_id)
                if status == "Completed":
                    break
                elif status == "Failed":
                    raise ValueError(extract_failure_error())
                elif status == "Cancelled":
                    print(
                        f"{icons.yellow_dot} The refresh of the '{dataset_name}' semantic model within the '{workspace_name}' workspace has been cancelled."
                    )
                    return

                time.sleep(3)

            print(
                f"{icons.green_dot} Refresh '{refresh_type}' of the '{dataset_name}' semantic model within the '{workspace_name}' workspace is complete."
            )

    final_output = refresh_and_trace_dataset(
        dataset=dataset,
        workspace=workspace,
        refresh_type=refresh_type,
        retry_count=retry_count,
        apply_refresh_policy=apply_refresh_policy,
        max_parallelism=max_parallelism,
        objects=objects,
        visualize=visualize,
        commit_mode=commit_mode,
    )

    return final_output


@log
def cancel_dataset_refresh(
    dataset: str | UUID,
    request_id: Optional[str] = None,
    workspace: Optional[str | UUID] = None,
):
    """
    Cancels the refresh of a semantic model which was executed via the `Enhanced Refresh API <https://learn.microsoft.com/power-bi/connect-data/asynchronous-refresh>`_

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    request_id : str, default=None
        The request id of a semantic model refresh.
        Defaults to finding the latest active refresh of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    rr = fabric.list_refresh_requests(dataset=dataset_id, workspace=workspace_id)
    rr_filt = rr[rr["Status"] == "Unknown"]

    if request_id is None:
        if len(rr_filt) == 0:
            raise ValueError(
                f"{icons.red_dot} There are no active Enhanced API refreshes of the '{dataset_name}' semantic model within the '{workspace_name}' workspace."
            )

        request_id = rr_filt["Request Id"].iloc[0]

    _base_api(
        request=f"/v1.0/myorg/groups/{workspace_id}/datasets/{dataset_id}/refreshes/{request_id}",
        method="delete",
    )
    print(
        f"{icons.green_dot} The '{request_id}' refresh request for the '{dataset_name}' semantic model within the '{workspace_name}' workspace has been cancelled."
    )


@log
def get_semantic_model_refresh_history(
    dataset: str | UUID,
    request_id: Optional[str] = None,
    workspace: Optional[str | UUID] = None,
) -> pd.DataFrame:
    """
    Obtains the semantic model refresh history (refreshes executed via the Enhanced Refresh API).

    This is a wrapper function for the following API: `Datasets - Get Refresh History In Group <https://learn.microsoft.com/rest/api/power-bi/datasets/get-refresh-history-in-group>`_.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    request_id : str, default=None
        The request id of a semantic model refresh.
        Defaults to None which resolves to showing all refresh requests for the given semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the semantic model refresh history.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)
    df = pd.DataFrame(
        columns=[
            "Request Id",
            "Refresh Type",
            "Start Time",
            "End Time",
            "Status",
            "Extended Status",
        ]
    )

    response = _base_api(
        request=f"/v1.0/myorg/groups/{workspace_id}/datasets/{dataset_id}/refreshes"
    )
    data = []

    for i in response.json().get("value", []):
        error = i.get("serviceExceptionJson")
        if error:
            error_json = json.loads(error)
        if request_id is None:
            new_data = {
                "Request Id": i.get("requestId"),
                "Refresh Type": i.get("refreshType"),
                "Start Time": i.get("startTime"),
                "End Time": i.get("endTime"),
                "Error Code": error_json.get("errorCode") if error else None,
                "Error Description": (
                    error_json.get("errorDescription") if error else None
                ),
                "Status": i.get("status"),
                "Extended Status": i.get("extendedStatus"),
                "Attempts": i.get("refreshAttempts"),
            }
            data.append(new_data)

        elif request_id == i.get("requestId"):
            for attempt in i.get("refreshAttempts", []):
                attempt_error = attempt.get("serviceExceptionJson")
                if attempt_error:
                    attempt_error_json = json.loads(attempt_error)
                new_data = {
                    "Request Id": i.get("requestId"),
                    "Refresh Type": i.get("refreshType"),
                    "Start Time": i.get("startTime"),
                    "End Time": i.get("endTime"),
                    "Error Code": error_json.get("errorCode") if error else None,
                    "Error Description": (
                        error_json.get("errorDescription") if error else None
                    ),
                    "Status": i.get("status"),
                    "Extended Status": i.get("extendedStatus"),
                    "Attempt Id": attempt.get("attemptId"),
                    "Attempt Start Time": attempt.get("startTime"),
                    "Attempt End Time": attempt.get("endTime"),
                    "Attempt Error Code": (
                        attempt_error_json.get("errorCode") if attempt_error else None
                    ),
                    "Attempt Error Description": (
                        attempt_error_json.get("errorDescription")
                        if attempt_error
                        else None
                    ),
                    "Type": attempt.get("type"),
                }
                data.append(new_data)

    if data:
        df = pd.DataFrame(data)

    # date_cols = ["Start Time", "End Time"]
    # df[date_cols] = df[date_cols].apply(pd.to_datetime)

    if "Attempt Id" in df.columns:
        column_map = {
            "Attempt Id": "int",
        }

        _update_dataframe_datatypes(dataframe=df, column_map=column_map)

    return df



from uuid import UUID
import sempy_labs._icons as icons
from typing import List, Optional
from sempy_labs._helper_functions import (
    _base_api,
)
from sempy._utils._log import log
from sempy_labs.graph._users import resolve_user_id


@log
def add_user_license(
    user: str | UUID, sku_id: UUID, disabled_plans: Optional[UUID | List[UUID]] = None
):
    """
    Assigns a license to a user.

    This is a wrapper function for the following API: `user: assignLicense <https://learn.microsoft.com/graph/api/user-assignlicense>`_.

    Service Principal Authentication is required (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    user : str | uuid.UUID
        The user ID or user principal name.
    sku_id : uuid.UUID
        The SKU ID of the license to assign.
    disabled_plans : Optional[uuid.UUID | List[uuid.UUID]], default=None
        A single service plan ID or a list of service plan IDs to disable within the assigned license.
    """

    user_id = resolve_user_id(user)

    payload = {
        "addLicenses": [
            {
                "skuId": sku_id,
            },
        ],
        "removeLicenses": [],
    }

    if disabled_plans:
        if isinstance(disabled_plans, str):
            disabled_plans = [disabled_plans]
        payload["addLicenses"][0]["disabledPlans"] = disabled_plans

    _base_api(
        request=f"users/{user_id}/assignLicense",
        client="graph",
        method="post",
        payload=payload,
    )

    print(
        f"{icons.green_dot} The '{sku_id}' license has been assigned to the user '{user}'."
    )


@log
def remove_user_license(user: str | UUID, sku_ids: UUID | List[UUID]):
    """
    Removes a license from a user.

    This is a wrapper function for the following API: `user: assignLicense <https://learn.microsoft.com/graph/api/user-assignlicense>`_.

    Service Principal Authentication is required (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    user : str | uuid.UUID
        The user ID or user principal name.
    sku_id : uuid.UUID
        The SKU ID of the license to remove.
    """

    user_id = resolve_user_id(user)

    if isinstance(sku_ids, str):
        sku_ids = [sku_ids]

    payload = {
        "addLicenses": [],
        "removeLicenses": sku_ids,
    }

    _base_api(
        request=f"users/{user_id}/assignLicense",
        client="graph",
        method="post",
        payload=payload,
    )

    print(
        f"{icons.green_dot} The '{', '.join([str(s) for s in sku_ids])}' license(s) have been removed from the user '{user}'."
    )



from .._helper_functions import (
    resolve_workspace_id,
    resolve_lakehouse_id,
    _base_api,
    _create_dataframe,
    _update_dataframe_datatypes,
)
import pandas as pd
from typing import Optional
from uuid import UUID
from sempy._utils._log import log


@log
def list_livy_sessions(
    lakehouse: Optional[str | UUID] = None, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Shows a list of livy sessions from the specified item identifier.

    This is a wrapper function for the following API: `Livy Sessions - List Livy Sessions <https://learn.microsoft.com/rest/api/fabric/lakehouse/livy-sessions/list-livy-sessions>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    lakehouse : str | uuid.UUID, default=None
        The Fabric lakehouse name or ID.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of livy sessions from the specified item identifier.
    """

    columns = {
        "Spark Application Id": "string",
        "State:": "string",
        "Livy Id": "string",
        "Origin": "string",
        "Attempt Number": "int",
        "Max Number Of Attempts": "int",
        "Livy Name": "string",
        "Submitter Id": "string",
        "Submitter Type": "string",
        "Item Workspace Id": "string",
        "Item Id": "string",
        "Item Reference Type": "string",
        "Item Name": "string",
        "Item Type": "string",
        "Job Type": "string",
        "Submitted Date Time": "str",
        "Start Date Time": "str",
        "End Date Time": "string",
        "Queued Duration Value": "int",
        "Queued Duration Time Unit": "string",
        "Running Duration Value": "int",
        "Running Duration Time Unit": "string",
        "Total Duration Value": "int",
        "Total Duration Time Unit": "string",
        "Job Instance Id": "string",
        "Creator Item Workspace Id": "string",
        "Creator Item Id": "string",
        "Creator Item Reference Type": "string",
        "Creator Item Name": "string",
        "Creator Item Type": "string",
        "Cancellation Reason": "string",
        "Capacity Id": "string",
        "Operation Name": "string",
        "Runtime Version": "string",
        "Livy Session Item Resource Uri": "string",
    }
    df = _create_dataframe(columns=columns)

    workspace_id = resolve_workspace_id(workspace)
    lakehouse_id = resolve_lakehouse_id(lakehouse, workspace_id)

    responses = _base_api(
        request=f"/v1/workspaces/{workspace_id}/lakehouses/{lakehouse_id}/livySessions",
        uses_pagination=True,
        client="fabric_sp",
    )

    rows = []
    for r in responses:
        for v in r.get("value", []):
            queued_duration = v.get("queuedDuration", {})
            running_duration = v.get("runningDuration", {})
            total_duration = v.get("totalDuration", {})
            rows.append(
                {
                    "Spark Application Id": v.get("sparkApplicationId"),
                    "State:": v.get("state"),
                    "Livy Id": v.get("livyId"),
                    "Origin": v.get("origin"),
                    "Attempt Number": v.get("attemptNumber"),
                    "Max Number Of Attempts": v.get("maxNumberOfAttempts"),
                    "Livy Name": v.get("livyName"),
                    "Submitter Id": v["submitter"].get("id"),
                    "Submitter Type": v["submitter"].get("type"),
                    "Item Workspace Id": v["item"].get("workspaceId"),
                    "Item Id": v["item"].get("itemId"),
                    "Item Reference Type": v["item"].get("referenceType"),
                    "Item Name": v.get("itemName"),
                    "Item Type": v.get("itemType"),
                    "Job Type": v.get("jobType"),
                    "Submitted Date Time": v.get("submittedDateTime"),
                    "Start Date Time": v.get("startDateTime"),
                    "End Date Time": v.get("endDateTime"),
                    "Queued Duration Value": queued_duration.get("value"),
                    "Queued Duration Time Unit": queued_duration.get("timeUnit"),
                    "Running Duration Value": running_duration.get("value"),
                    "Running Duration Time Unit": running_duration.get("timeUnit"),
                    "Total Duration Value": total_duration.get("value"),
                    "Total Duration Time Unit": total_duration.get("timeUnit"),
                    "Job Instance Id": v.get("jobInstanceId"),
                    "Creator Item Workspace Id": v["creatorItem"].get("workspaceId"),
                    "Creator Item Id": v["creatorItem"].get("itemId"),
                    "Creator Item Reference Type": v["creatorItem"].get(
                        "referenceType"
                    ),
                    "Creator Item Name": v.get("creatorItemName"),
                    "Creator Item Type": v.get("creatorItemType"),
                    "Cancellation Reason": v.get("cancellationReason"),
                    "Capacity Id": v.get("capacityId"),
                    "Operation Name": v.get("operationName"),
                    "Runtime Version": v.get("runtimeVersion"),
                    "Livy Session Item Resource Uri": v.get(
                        "livySessionItemResourceUri"
                    ),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))
        _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df



import pandas as pd
from typing import Optional
from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    resolve_workspace_id,
    _decode_b64,
    _base_api,
    resolve_item_id,
    _create_dataframe,
    delete_item,
    create_item,
)
from uuid import UUID
from sempy._utils._log import log


@log
def list_data_pipelines(workspace: Optional[str | UUID] = None) -> pd.DataFrame:
    """
    Shows the data pipelines within a workspace.

    This is a wrapper function for the following API: `Items - List Data Pipelines <https://learn.microsoft.com/rest/api/fabric/datapipeline/items/list-data-pipelines>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the data pipelines within a workspace.
    """

    columns = {
        "Data Pipeline Name": "string",
        "Data Pipeline ID": "string",
        "Description": "string",
    }
    df = _create_dataframe(columns=columns)

    workspace_id = resolve_workspace_id(workspace)

    responses = _base_api(
        request=f"/v1/workspaces/{workspace_id}/dataPipelines",
        uses_pagination=True,
        client="fabric_sp",
    )

    rows = []
    for r in responses:
        for v in r.get("value", []):
            rows.append(
                {
                    "Data Pipeline Name": v.get("displayName"),
                    "Data Pipeline ID": v.get("id"),
                    "Description": v.get("description"),
                }
            )
    if rows:
        df = pd.DataFrame(rows, columns=columns.keys())

    return df


@log
def create_data_pipeline(
    name: str, description: Optional[str] = None, workspace: Optional[str | UUID] = None
):
    """
    Creates a Fabric data pipeline.

    This is a wrapper function for the following API: `Items - Create Data Pipeline <https://learn.microsoft.com/rest/api/fabric/datapipeline/items/create-data-pipeline>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    name: str
        Name of the data pipeline.
    description : str, default=None
        A description of the environment.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    create_item(
        name=name, description=description, type="DataPipeline", workspace=workspace
    )


@log
def delete_data_pipeline(name: str | UUID, workspace: Optional[str | UUID] = None):
    """
    Deletes a Fabric data pipeline.

    This is a wrapper function for the following API: `Items - Delete Data Pipeline <https://learn.microsoft.com/rest/api/fabric/datapipeline/items/delete-data-pipeline>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    name: str | uuid.UUID
        Name or ID of the data pipeline.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    delete_item(item=name, type="DataPipeline", workspace=workspace)


@log
def get_data_pipeline_definition(
    name: str | UUID, workspace: Optional[str | UUID] = None, decode: bool = True
) -> dict | pd.DataFrame:
    """
    Obtains the definition of a data pipeline.

    Parameters
    ----------
    name : str or uuid.UUID
        The name or ID of the data pipeline.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    decode : bool, default=True
        decode : bool, default=True
        If True, decodes the data pipeline definition file into .json format.
        If False, obtains the data pipeline definition file a pandas DataFrame format.

    Returns
    -------
    dict | pandas.DataFrame
        A pandas dataframe showing the data pipelines within a workspace.
    """

    workspace_id = resolve_workspace_id(workspace)

    item_id = resolve_item_id(item=name, type="DataPipeline", workspace=workspace)
    result = _base_api(
        request=f"/v1/workspaces/{workspace_id}/dataPipelines/{item_id}/getDefinition",
        method="post",
        lro_return_json=True,
        status_codes=None,
        client="fabric_sp",
    )
    df = pd.json_normalize(result["definition"]["parts"])

    if not decode:
        return df
    content = df[df["path"] == "pipeline-content.json"]
    payload = content["payload"].iloc[0]

    return _decode_b64(payload)



import pandas as pd
from uuid import UUID
import sempy_labs._icons as icons
from typing import Optional, Tuple
from sempy._utils._log import log
from sempy_labs._helper_functions import (
    _base_api,
    _build_url,
    _create_dataframe,
    _update_dataframe_datatypes,
    _is_valid_uuid,
    get_capacity_id,
)


@log
def patch_capacity(capacity: str | UUID, tenant_key_id: UUID):
    """
    Changes specific capacity information. Currently, this API call only supports changing the capacity's encryption key.

    This is a wrapper function for the following API: `Admin - Patch Capacity As Admin <https://learn.microsoft.com/rest/api/power-bi/admin/patch-capacity-as-admin>`_.

    Parameters
    ----------
    capacity : str | uuid.UUID
        The name or ID of the capacity.
    tenant_key_id : str
        The ID of the encryption key.
    """

    (capacity_name, capacity_id) = _resolve_capacity_name_and_id(capacity)

    payload = {
        "tenantKeyId": tenant_key_id,
    }

    _base_api(
        request=f"/v1.0/myorg/admin/capacities/{capacity_id}",
        method="patch",
        payload=payload,
    )

    print(
        f"{icons.green_dot} The capacity '{capacity_name}' has been successfully patched."
    )


@log
def _resolve_capacity_name_and_id(
    capacity: str | UUID,
) -> Tuple[str, UUID]:

    dfC = list_capacities(capacity=capacity)
    if dfC.empty:
        raise ValueError(f"{icons.red_dot} The '{capacity}' capacity was not found.")

    capacity_name = dfC["Capacity Name"].iloc[0]
    capacity_id = dfC["Capacity Id"].iloc[0]

    return capacity_name, capacity_id


@log
def _resolve_capacity_id(
    capacity: str | UUID,
) -> UUID:

    if _is_valid_uuid(capacity):
        capacity_id = capacity
    else:
        dfC = list_capacities(capacity=capacity)
        if dfC.empty:
            raise ValueError(
                f"{icons.red_dot} The '{capacity}' capacity was not found."
            )

        capacity_id = dfC["Capacity Id"].iloc[0]

    return capacity_id


@log
def _list_capacities_meta() -> pd.DataFrame:
    """
    Shows the a list of capacities and their properties. This function is the admin version.

    This is a wrapper function for the following API: `Admin - Get Capacities As Admin <https://learn.microsoft.com/rest/api/power-bi/admin/get-capacities-as-admin>`_.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the capacities and their properties
    """

    columns = {
        "Capacity Id": "string",
        "Capacity Name": "string",
        "Sku": "string",
        "Region": "string",
        "State": "string",
        "Admins": "string",
    }
    df = _create_dataframe(columns=columns)

    responses = _base_api(
        request="/v1.0/myorg/admin/capacities", client="fabric_sp", uses_pagination=True
    )

    rows = []
    for r in responses:
        for i in r.get("value", []):
            rows.append(
                {
                    "Capacity Id": i.get("id").lower(),
                    "Capacity Name": i.get("displayName"),
                    "Sku": i.get("sku"),
                    "Region": i.get("region"),
                    "State": i.get("state"),
                    "Admins": [i.get("admins", [])],
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df


@log
def get_capacity_assignment_status(
    workspace: Optional[str | UUID] = None,
) -> pd.DataFrame:
    """
    Gets the status of the assignment-to-capacity operation for the specified workspace.

    This is a wrapper function for the following API: `Capacities - Groups CapacityAssignmentStatus <https://learn.microsoft.com/rest/api/power-bi/capacities/groups-capacity-assignment-status>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the status of the assignment-to-capacity operation for the specified workspace.
    """
    from sempy_labs.admin._basic_functions import _resolve_workspace_name_and_id

    (workspace_name, workspace_id) = _resolve_workspace_name_and_id(workspace)

    columns = {
        "Status": "string",
        "Activity Id": "string",
        "Start Time": "datetime",
        "End Time": "datetime",
        "Capacity Id": "string",
        "Capacity Name": "string",
    }
    df = _create_dataframe(columns=columns)

    response = _base_api(
        request=f"/v1.0/myorg/groups/{workspace_id}/CapacityAssignmentStatus",
        client="fabric_sp",
    )
    v = response.json()
    capacity_id = v.get("capacityId")

    (capacity_name, capacity_id) = _resolve_capacity_name_and_id(capacity=capacity_id)

    new_data = {
        "Status": v.get("status"),
        "Activity Id": v.get("activityId"),
        "Start Time": v.get("startTime"),
        "End Time": v.get("endTime"),
        "Capacity Id": capacity_id,
        "Capacity Name": capacity_name,
    }

    df = pd.concat([df, pd.DataFrame([new_data])], ignore_index=True)

    _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


@log
def get_capacity_state(capacity: Optional[str | UUID] = None):
    """
    Gets the state of a capacity.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    capacity : str | uuid.UUID, default=None
        The capacity name or ID.
        Defaults to None which resolves to the capacity of the attached lakehouse
        or if no lakehouse is attached, resolves to the workspace of the notebook.

    Returns
    -------
    str
        The capacity state.
    """

    df = list_capacities()

    if capacity is None:
        capacity = get_capacity_id()
    if _is_valid_uuid(capacity):
        df_filt = df[df["Capacity Id"] == capacity]
    else:
        df_filt = df[df["Capacity Name"] == capacity]

    if df_filt.empty:
        raise ValueError(f"{icons.red_dot} The capacity '{capacity}' was not found.")

    return df_filt["State"].iloc[0]


@log
def list_capacities(
    capacity: Optional[str | UUID] = None,
    include_tenant_key: bool = False,
) -> pd.DataFrame:
    """
    Shows the a list of capacities and their properties.

    This is a wrapper function for the following API: `Admin - Get Capacities As Admin <https://learn.microsoft.com/rest/api/power-bi/admin/get-capacities-as-admin>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    capacity : str | uuid.UUID, default=None
        The capacity name or ID to filter. If None, all capacities are returned.
    include_tenant_key : bool, default=False
        If True, obtains the `tenant key <https://learn.microsoft.com/rest/api/power-bi/admin/get-capacities-as-admin#example-with-expand-on-tenant-key>`_ properties.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the capacities and their properties.
    """

    columns = {
        "Capacity Id": "string",
        "Capacity Name": "string",
        "Sku": "string",
        "Region": "string",
        "State": "string",
        "Admins": "list",
        "Users": "list",
    }
    if include_tenant_key:
        columns.update(
            {
                "Tenant Key Id": "string",
                "Tenant Key Name": "string",
            }
        )
    df = _create_dataframe(columns=columns)

    url = "/v1.0/myorg/admin/capacities"
    if include_tenant_key:
        url += "?$expand=tenantKey"

    responses = _base_api(request=url, client="fabric_sp", uses_pagination=True)

    rows = []
    for r in responses:
        for i in r.get("value", []):
            row = {
                "Capacity Id": i.get("id", "").lower(),
                "Capacity Name": i.get("displayName"),
                "Sku": i.get("sku"),
                "Region": i.get("region"),
                "State": i.get("state"),
                "Admins": i.get("admins", []),
                "Users": i.get("users", []),
            }

            if include_tenant_key:
                tenant_key = i.get("tenantKey") or {}
                row.update(
                    {
                        "Tenant Key Id": tenant_key.get("id"),
                        "Tenant Key Name": tenant_key.get("name"),
                    }
                )

            rows.append(row)

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    if capacity is not None:
        if _is_valid_uuid(capacity):
            df = df[df["Capacity Id"] == capacity.lower()]
        else:
            df = df[df["Capacity Name"] == capacity]

    return df


@log
def list_capacity_users(capacity: str | UUID) -> pd.DataFrame:
    """
    Shows a list of users that have access to the specified workspace.

    This is a wrapper function for the following API: `Admin - Capacities GetCapacityUsersAsAdmin <https://learn.microsoft.com/rest/api/power-bi/admin/capacities-get-capacity-users-as-admin>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    capacity : str | uuid.UUID
        The name or ID of the capacity.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of users that have access to the specified workspace.
    """

    (_, capacity_id) = _resolve_capacity_name_and_id(capacity)

    columns = {
        "User Name": "string",
        "Email Address": "string",
        "Capacity User Access Right": "string",
        "Identifier": "string",
        "Graph Id": "string",
        "Principal Type": "string",
        "User Type": "string",
        "Profile": "string",
    }

    df = _create_dataframe(columns=columns)

    response = _base_api(
        request=f"/v1.0/myorg/admin/capacities/{capacity_id}/users", client="fabric_sp"
    )

    rows = []
    for v in response.json().get("value", []):
        rows.append(
            {
                "User Name": v.get("displayName"),
                "Email Address": v.get("emailAddress"),
                "Capacity User Access Right": v.get("capacityUserAccessRight"),
                "Identifier": v.get("identifier"),
                "Graph Id": v.get("graphId"),
                "Principal Type": v.get("principalType"),
                "User Type": v.get("userType"),
                "Profile": v.get("profile"),
            }
        )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))
        _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


@log
def get_refreshables(
    top: Optional[int] = None,
    expand: Optional[str] = None,
    filter: Optional[str] = None,
    skip: Optional[int] = None,
    capacity: Optional[str | UUID] = None,
) -> pd.DataFrame | dict:
    """
    Returns a list of refreshables for the organization within a capacity.

    Power BI retains a seven-day refresh history for each dataset, up to a maximum of sixty refreshes.

    This is a wrapper function for the following API: `Admin - Get Refreshables <https://learn.microsoft.com/rest/api/power-bi/admin/get-refreshables>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    top : int, default=None
        Returns only the first n results.
    expand : str, default=None
        Accepts a comma-separated list of data types, which will be expanded inline in the response. Supports capacities and groups.
    filter : str, default=None
        Returns a subset of a results based on Odata filter query parameter condition.
    skip : int, default=None
        Skips the first n results. Use with top to fetch results beyond the first 1000.
    capacity : str | uuid.UUID, default=None
        The capacity name or ID to filter. If None, all capacities are returned.

    Returns
    -------
    pandas.DataFrame
        Returns a list of refreshables for the organization within a capacity.
    """

    columns = {
        "Workspace Id": "string",
        "Workspace Name": "string",
        "Item Id": "string",
        "Item Name": "string",
        "Item Kind": "string",
        "Capacity Id": "string",
        "Capacity Name": "string",
        "Capacity SKU": "string",
        "Refresh Count": "int",
        "Refresh Failures": "int",
        "Average Duration": "float",
        "Median Duration": "float",
        "Refreshes Per Day": "int",
        "Refresh Type": "string",
        "Start Time": "string",
        "End Time": "string",
        "Status": "string",
        "Request Id": "string",
        "Service Exception Json": "string",
        "Extended Status": "dict",
        "Refresh Attempts": "list",
        "Refresh Schedule Days": "list",
        "Refresh Schedule Times": "list",
        "Refresh Schedule Enabled": "bool",
        "Refresh Schedule Local Timezone Id": "string",
        "Refresh Schedule Notify Option": "string",
        "Configured By": "list",
    }

    df = _create_dataframe(columns=columns)

    params = {}
    url = (
        "/v1.0/myorg/admin/capacities/refreshables"
        if capacity is None
        else f"/v1.0/myorg/admin/capacities/{_resolve_capacity_id(capacity=capacity)}/refreshables"
    )

    if top is not None:
        params["$top"] = top

    if expand is not None:
        params["$expand"] = expand

    if filter is not None:
        params["$filter"] = filter

    if skip is not None:
        params["$skip"] = skip

    url = _build_url(url, params)

    responses = _base_api(request=url, client="fabric_sp")

    rows = []
    for i in responses.json().get("value", []):
        last_refresh = i.get("lastRefresh", {})
        refresh_schedule = i.get("refreshSchedule", {})
        rows.append(
            {
                "Workspace Id": i.get("group", {}).get("id"),
                "Workspace Name": i.get("group", {}).get("name"),
                "Item Id": i.get("id"),
                "Item Name": i.get("name"),
                "Item Kind": i.get("kind"),
                "Capacity Id": (
                    i.get("capacity", {}).get("id").lower()
                    if i.get("capacity", {}).get("id")
                    else None
                ),
                "Capacity Name": i.get("capacity", {}).get("displayName"),
                "Capacity SKU": i.get("capacity", {}).get("sku"),
                "Refresh Count": i.get("refreshCount", 0),
                "Refresh Failures": i.get("refreshFailures", 0),
                "Average Duration": i.get("averageDuration", 0),
                "Median Duration": i.get("medianDuration", 0),
                "Refreshes Per Day": i.get("refreshesPerDay", 0),
                "Refresh Type": last_refresh.get("refreshType"),
                "Start Time": last_refresh.get("startTime"),
                "End Time": last_refresh.get("endTime"),
                "Status": last_refresh.get("status"),
                "Request Id": last_refresh.get("requestId"),
                "Service Exception Json": last_refresh.get("serviceExceptionJson"),
                "Extended Status": last_refresh.get("extendedStatus"),
                "Refresh Attempts": last_refresh.get("refreshAttempts"),
                "Refresh Schedule Days": refresh_schedule.get("days"),
                "Refresh Schedule Times": refresh_schedule.get("times"),
                "Refresh Schedule Enabled": refresh_schedule.get("enabled"),
                "Refresh Schedule Local Timezone Id": refresh_schedule.get(
                    "localTimeZoneId"
                ),
                "Refresh Schedule Notify Option": refresh_schedule.get("notifyOption"),
                "Configured By": i.get("configuredBy"),
            }
        )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))
        _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df



from ._refresh_catalog_metadata import (
    refresh_catalog_metadata,
)
from ._discover import (
    discover_catalogs,
    discover_schemas,
    discover_tables,
)

__all__ = [
    "refresh_catalog_metadata",
    "discover_catalogs",
    "discover_schemas",
    "discover_tables",
]



item_types = {
    "Dashboard": ["Dashboard", "dashboards"],
    "DataPipeline": ["Data Pipeline", "dataPipelines", "pipeline-content.json"],
    "Datamart": ["Datamart", "datamarts"],
    "Environment": ["Environment", "environments"],
    "Eventhouse": ["Eventhouse", "eventhouses", "EventhouseProperties.json"],
    "Eventstream": ["Eventstream", "eventstreams", "eventstream.json"],
    "GraphQLApi": ["GraphQL Api", "GraphQLApis"],
    "KQLDashboard": ["KQL Dashboard", "kqlDashboards", "RealTimeDashboard.json"],
    "KQLDatabase": [
        "KQL Database",
        "kqlDatabases",
    ],  # "DatabaseProperties.json", "DatabaseSchema.kql"
    "KQLQueryset": ["KQL Queryset", "kqlQuerysets", "RealTimeQueryset.json"],
    "Lakehouse": ["Lakehouse", "lakehouses"],
    "MLExperiment": ["ML Experiment", "mlExperiments"],
    "MLModel": ["ML Model", "mlModels"],
    "MirroredDatabase": [
        "Mirrored Database",
        "mirroredDatabases",
        "mirroredDatabase.json",
    ],
    "MirroredWarehouse": ["Mirrored Warehouse", "mirroredWarehouses"],
    "MountedDataFactory": [
        "Mounted Data Factory",
        "mountedDataFactories",
        "mountedDataFactory-content.json",
    ],
    "Notebook": ["Notebook", "notebooks"],
    "PaginatedReport": ["Paginated Report", "paginatedReports"],
    "Reflex": ["Reflex", "reflexes", "ReflexEntities.json"],
    "Report": ["Report", "reports", "report.json"],
    "SQLDatabase": ["SQL Database", "sqlDatabases"],
    "SQLEndpoint": ["SQL Endpoint", "sqlEndpoints"],
    "SemanticModel": ["Semantic Model", "semanticModels", "model.bim"],
    "SparkJobDefinition": [
        "Spark Job Definition",
        "sparkJobDefinitions",
        "SparkJobDefinitionV1.json",
    ],
    "Warehouse": ["Warehouse", "warehouses"],
}


items = {
    "CopyJob": "copyJobs",
    "Dataflow": "dataflows",
    "Eventhouse": "eventhouses",
    "GraphQLApi": "GraphQLApis",
    "Report": "reports",
    "SemanticModel": "semanticModels",
    # "Environment": "environments",
    "KQLDatabase": "kqlDatabases",
    "KQLDashboard": "kqlDashboards",
    "KQLQueryset": "kqlQuerysets",
    "DataPipeline": "dataPipelines",
    "Notebook": "notebooks",
    "SparkJobDefinition": "sparkJobDefinitions",
    "Eventstream": "eventstreams",
    "MirroredWarehouse": "mirroredWarehouses",
    "MirroredDatabase": "mirroredDatabases",
    "MountedDataFactory": "mountedDataFactories",
    "VariableLibrary": "variableLibraries",
    "ApacheAirFlowJob": "ApacheAirflowJobs",
    "WarehouseSnapshot": "warehousesnapshots",
    "DigitalTwinBuilder": "digitaltwinbuilders",
    "DigitalTwinBuilderFlow": "DigitalTwinBuilderFlows",
    "MirroredAzureDatabricksCatalog": "mirroredAzureDatabricksCatalogs",
    "Map": "Maps",
    "AnomalyDetector": "anomalydetectors",
}



import sempy.fabric as fabric
import json
import os
import time
from sempy_labs._helper_functions import (
    generate_embedded_filter,
    resolve_workspace_name_and_id,
    _base_api,
    _mount,
)
from typing import Optional
from sempy._utils._log import log
import sempy_labs._icons as icons
from uuid import UUID
from sempy_labs.report._report_functions import (
    list_report_visuals,
    list_report_pages,
)


@log
def export_report(
    report: str,
    export_format: str,
    file_name: Optional[str] = None,
    bookmark_name: Optional[str] = None,
    page_name: Optional[str] = None,
    visual_name: Optional[str] = None,
    report_filter: Optional[str] = None,
    workspace: Optional[str | UUID] = None,
    lakehouse: Optional[str | UUID] = None,
    lakehouse_workspace: Optional[str | UUID] = None,
):
    """
    Exports a Power BI report to a file in your lakehouse.

    This is a wrapper function for the following APIs: `Reports - Export To File In Group <https://learn.microsoft.com/rest/api/power-bi/reports/export-to-file-in-group>`_, `Reports - Get Export To File Status In Group <https://learn.microsoft.com/rest/api/power-bi/reports/get-export-to-file-status-in-group>`_, `Reports - Get File Of Export To File In Group <https://learn.microsoft.com/rest/api/power-bi/reports/get-file-of-export-to-file-in-group>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    report : str
        Name of the Power BI report.
    export_format : str
        The format in which to export the report. For image formats, enter the file extension in this parameter, not 'IMAGE'.
        `Valid formats <https://learn.microsoft.com/rest/api/power-bi/reports/export-to-file-in-group#fileformat>`_
    file_name : str, default=None
        The name of the file to be saved within the lakehouse. Do not include the file extension. Defaults ot the reportName parameter value.
    bookmark_name : str, default=None
        The name (GUID) of a bookmark within the report.
    page_name : str, default=None
        The name (GUID) of the report page.
    visual_name : str, default=None
        The name (GUID) of a visual. If you specify this parameter you must also specify the page_name parameter.
    report_filter : str, default=None
        A report filter to be applied when exporting the report. Syntax is user-friendly. See above for examples.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    lakehouse : str | uuid.UUID, default=None
        The Fabric lakehouse name or ID. This will be the lakehouse to which the export of the report is saved.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    lakehouse_workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID used by the lakehouse.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    if isinstance(page_name, str):
        page_name = [page_name]
    if isinstance(visual_name, str):
        visual_name = [visual_name]

    if bookmark_name is not None and (page_name is not None or visual_name is not None):
        raise ValueError(
            f"{icons.red_dot} If the 'bookmark_name' parameter is set, the 'page_name' and 'visual_name' parameters must not be set."
        )

    if visual_name is not None and page_name is None:
        raise ValueError(
            f"{icons.red_dot} If the 'visual_name' parameter is set, the 'page_name' parameter must be set."
        )

    valid_formats = {
        "ACCESSIBLEPDF": ".pdf",
        "CSV": ".csv",
        "DOCX": ".docx",
        "MHTML": ".mhtml",
        "PDF": ".pdf",
        "PNG": ".png",
        "PPTX": ".pptx",
        "XLSX": ".xlsx",
        "XML": ".xml",
        "BMP": ".bmp",
        "EMF": ".emf",
        "GIF": ".gif",
        "JPEG": ".jpeg",
        "TIFF": ".tiff",
    }

    export_format = export_format.upper()
    file_ext = valid_formats.get(export_format)
    if file_ext is None:
        raise ValueError(
            f"{icons.red_dot} The '{export_format}' format is not a valid format for exporting Power BI reports. Please enter a valid format. Options: {valid_formats}"
        )

    if file_name is None:
        file_name = f"{report}{file_ext}"
    else:
        file_name = f"{file_name}{file_ext}"

    dfI = fabric.list_items(workspace=workspace)
    dfI_filt = dfI[
        (dfI["Type"].isin(["Report", "PaginatedReport"]))
        & (dfI["Display Name"] == report)
    ]

    if dfI_filt.empty:
        raise ValueError(
            f"{icons.red_dot} The '{report}' report does not exist in the '{workspace_name}' workspace."
        )

    report_type = dfI_filt["Type"].iloc[0]

    # Limitations
    pbiOnly = ["PNG"]
    paginatedOnly = [
        "ACCESSIBLEPDF",
        "CSV",
        "DOCX",
        "BMP",
        "EMF",
        "GIF",
        "JPEG",
        "TIFF",
        "MHTML",
        "XLSX",
        "XML",
    ]

    if report_type == "Report" and export_format in paginatedOnly:
        raise ValueError(
            f"{icons.red_dot} The '{export_format}' format is only supported for paginated reports."
        )

    if report_type == "PaginatedReport" and export_format in pbiOnly:
        raise ValueError(
            f"{icons.red_dot} The '{export_format}' format is only supported for Power BI reports."
        )

    if report_type == "PaginatedReport" and (
        bookmark_name is not None or page_name is not None or visual_name is not None
    ):
        raise ValueError(
            f"{icons.red_dot} Export for paginated reports does not support bookmarks/pages/visuals. Those parameters must not be set for paginated reports."
        )

    reportId = dfI_filt["Id"].iloc[0]

    if (
        export_format in ["BMP", "EMF", "GIF", "JPEG", "TIFF"]
        and report_type == "PaginatedReport"
    ):
        request_body = {
            "format": "IMAGE",
            "paginatedReportConfiguration": {
                "formatSettings": {"OutputFormat": export_format.lower()}
            },
        }
    elif bookmark_name is None and page_name is None and visual_name is None:
        request_body = {"format": export_format}
    elif bookmark_name is not None:
        if report_type == "Report":
            request_body = {
                "format": export_format,
                "powerBIReportConfiguration": {
                    "defaultBookmark": {"name": bookmark_name}
                },
            }
    elif page_name is not None and visual_name is None:
        if report_type == "Report":
            request_body = {"format": export_format, "powerBIReportConfiguration": {}}

            request_body["powerBIReportConfiguration"]["pages"] = []
            dfPage = list_report_pages(report=report, workspace=workspace_id)

            for page in page_name:
                dfPage_filt = dfPage[dfPage["Page ID"] == page]
                if len(dfPage_filt) == 0:
                    raise ValueError(
                        f"{icons.red_dot} The '{page}' page does not exist in the '{report}' report within the '{workspace_name}' workspace."
                    )

                page_dict = {"pageName": page}
                request_body["powerBIReportConfiguration"]["pages"].append(page_dict)

    elif page_name is not None and visual_name is not None:
        if len(page_name) != len(visual_name):
            raise ValueError(
                f"{icons.red_dot} Each 'visual_name' must map to a single 'page_name'."
            )

        if report_type == "Report":
            request_body = {"format": export_format, "powerBIReportConfiguration": {}}

            request_body["powerBIReportConfiguration"]["pages"] = []
            dfVisual = list_report_visuals(report=report, workspace=workspace_id)
            a = 0
            for page in page_name:
                visual = visual_name[a]

                dfVisual_filt = dfVisual[
                    (dfVisual["Page ID"] == page) & (dfVisual["Visual ID"] == visual)
                ]
                if len(dfVisual_filt) == 0:
                    raise ValueError(
                        f"{icons.red_dot} The '{visual}' visual does not exist on the '{page}' in the '{report}' report within the '{workspace_name}' workspace."
                    )

                page_dict = {"pageName": page, "visualName": visual}
                request_body["powerBIReportConfiguration"]["pages"].append(page_dict)
                a += 1

    # Transform and add report filter if it is specified
    if report_filter is not None and report_type == "Report":
        reportFilter = generate_embedded_filter(filter=report_filter)
        report_level_filter = {"filter": reportFilter}

        if "powerBIReportConfiguration" not in request_body:
            request_body["powerBIReportConfiguration"] = {}
        request_body["powerBIReportConfiguration"]["reportLevelFilters"] = [
            report_level_filter
        ]

    base_url = f"/v1.0/myorg/groups/{workspace_id}/reports/{reportId}"
    response = _base_api(
        request=f"{base_url}/ExportTo",
        method="post",
        payload=request_body,
        status_codes=202,
        client="fabric_sp",
    )
    export_id = json.loads(response.content).get("id")

    get_status_url = f"{base_url}/exports/{export_id}"
    response = _base_api(
        request=get_status_url, status_codes=[200, 202], client="fabric_sp"
    )
    response_body = json.loads(response.content)
    while response_body["status"] not in ["Succeeded", "Failed"]:
        time.sleep(3)
        response = _base_api(
            request=get_status_url, status_codes=[200, 202], client="fabric_sp"
        )
        response_body = json.loads(response.content)
    if response_body["status"] == "Failed":
        raise ValueError(
            f"{icons.red_dot} The export for the '{report}' report within the '{workspace_name}' workspace in the '{export_format}' format has failed."
        )
    else:
        response = _base_api(request=f"{get_status_url}/file", client="fabric_sp")
        print(
            f"{icons.in_progress} Saving the '{export_format}' export for the '{report}' report within the '{workspace_name}' workspace to the lakehouse..."
        )

        local_path = _mount(lakehouse=lakehouse, workspace=lakehouse_workspace)
        folder_path = f"{local_path}/Files"
        file_path = os.path.join(folder_path, file_name)

        with open(file_path, "wb") as export_file:
            export_file.write(response.content)
        print(
            f"{icons.green_dot} The '{export_format}' export for the '{report}' report within the '{workspace_name}' workspace has been saved '{file_name}' in the '{lakehouse}' within the '{lakehouse_workspace}' workspace."
        )



import sempy.fabric as fabric
import pandas as pd
from .tom import connect_semantic_model
from ._generate_semantic_model import create_blank_semantic_model
from .directlake._generate_shared_expression import generate_shared_expression
from typing import Optional
from sempy._utils._log import log


@log
def model_auto_build(
    dataset: str,
    file_path: str,
    workspace: Optional[str] = None,
    lakehouse: Optional[str] = None,
    lakehouse_workspace: Optional[str] = None,
):
    """
    Dynamically generates a semantic model based on an Excel file template.

    Parameters
    ----------
    dataset : str
        Name of the semantic model.
    file_path : str
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    lakehouse : str, default=None
        The Fabric lakehouse used by the Direct Lake semantic model.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    lakehouse_workspace : str, default=None
        The Fabric workspace used by the lakehouse.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    workspace = fabric.resolve_workspace_name(workspace)

    if lakehouse_workspace is None:
        lakehouse_workspace = workspace

    sheets = [
        "Model",
        "Tables",
        "Measures",
        "Columns",
        "Roles",
        "Hierarchies",
        "Relationships",
    ]

    create_blank_semantic_model(dataset=dataset, workspace=workspace)

    with connect_semantic_model(
        dataset=dataset, workspace=workspace, readonly=False
    ) as tom:

        # DL Only
        expr = generate_shared_expression(
            item_name=lakehouse, item_type="Lakehouse", workspace=lakehouse_workspace
        )
        tom.add_expression(name="DatbaseQuery", expression=expr)

        for sheet in sheets:
            df = pd.read_excel(file_path, sheet_name=sheet)

            if sheet == "Tables":
                for i, r in df.iterrows():
                    tName = r["Table Name"]
                    desc = r["Description"]
                    dc = r["Data Category"]
                    mode = r["Mode"]
                    hidden = bool(r["Hidden"])

                    tom.add_table(
                        name=tName, description=desc, data_category=dc, hidden=hidden
                    )
                    if mode == "DirectLake":
                        tom.add_entity_partition(table_name=tName, entity_name=tName)
            elif sheet == "Columns":
                for i, r in df.iterrows():
                    tName = r["Table Name"]
                    cName = r["Column Name"]
                    scName = r["Source Column"]
                    dataType = r["Data Type"]
                    hidden = bool(r["Hidden"])
                    key = bool(r["Key"])
                    if dataType == "Integer":
                        dataType = "Int64"
                    desc = r["Description"]

                    tom.add_data_column(
                        table_name=tName,
                        column_name=cName,
                        source_column=scName,
                        data_type=dataType,
                        description=desc,
                        hidden=hidden,
                        key=key,
                    )
            elif sheet == "Measures":
                for i, r in df.iterrows():
                    tName = r["Table Name"]
                    mName = r["Measure Name"]
                    expr = r["Expression"]
                    desc = r["Description"]
                    format = r["Format String"]
                    hidden = bool(r["Hidden"])

                    tom.add_measure(
                        table_name=tName,
                        measure_name=mName,
                        expression=expr,
                        format_string=format,
                        description=desc,
                        hidden=hidden,
                    )
            elif sheet == "Relationships":
                for i, r in df.iterrows():
                    fromTable = r["From Table"]
                    fromColumn = r["From Column"]
                    toTable = r["To Table"]
                    toColumn = r["To Column"]
                    fromCard = r["From Cardinality"]
                    toCard = r["To Cardinality"]

                    tom.add_relationship(
                        from_table=fromTable,
                        from_column=fromColumn,
                        to_table=toTable,
                        to_column=toColumn,
                        from_cardinality=fromCard,
                        to_cardinality=toCard,
                    )
            elif sheet == "Roles":
                print("hi")
            elif sheet == "Hierarchies":
                print("hi")



from ._get_lakehouse_columns import (
    get_lakehouse_columns,
)
from ._get_lakehouse_tables import (
    get_lakehouse_tables,
)
from ._lakehouse import (
    lakehouse_attached,
    optimize_lakehouse_tables,
    vacuum_lakehouse_tables,
    run_table_maintenance,
)
from ._shortcuts import (
    # create_shortcut,
    create_shortcut_onelake,
    delete_shortcut,
    reset_shortcut_cache,
    list_shortcuts,
)
from ._blobs import (
    recover_lakehouse_object,
    list_blobs,
)
from ._livy_sessions import (
    list_livy_sessions,
)
from ._helper import (
    is_v_ordered,
    delete_lakehouse,
    update_lakehouse,
    load_table,
)
from ._materialized_lake_views import (
    refresh_materialized_lake_views,
)

__all__ = [
    "get_lakehouse_columns",
    "get_lakehouse_tables",
    "lakehouse_attached",
    "optimize_lakehouse_tables",
    # create_shortcut,
    "create_shortcut_onelake",
    "delete_shortcut",
    "vacuum_lakehouse_tables",
    "reset_shortcut_cache",
    "run_table_maintenance",
    "list_shortcuts",
    "recover_lakehouse_object",
    "list_blobs",
    "list_livy_sessions",
    "is_v_ordered",
    "delete_lakehouse",
    "update_lakehouse",
    "load_table",
    "refresh_materialized_lake_views",
]
